# Volatility
# Copyright (c) 2008-2015 Volatility Foundation
#
# This file is part of Volatility.
#
# Volatility is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License Version 2 as
# published by the Free Software Foundation.  You may not use, modify or
# distribute this program under any other version of the GNU General
# Public License.
#
# Volatility is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Volatility.  If not, see <http://www.gnu.org/licenses/>.
#

"""
@author:       The Volatility Foundation
@license:      GNU General Public License 2.0
@contact:      awalters@4tphi.net

This file provides support for Windows 10.
"""

import volatility.plugins.overlays.windows.windows as windows
import volatility.obj as obj
import volatility.win32.tasks as tasks 
import volatility.debug as debug
import volatility.plugins.overlays.windows.win8 as win8
from volatility.win32.rawreg import KEY_FLAGS

try:
    import distorm3
    has_distorm = True
except ImportError:
    has_distorm = False

class _HMAP_ENTRY(obj.CType):

    @property
    def BlockAddress(self):
        return self.PermanentBinAddress & 0xFFFFFFFFFFF0

class _CM_KEY_BODY(windows._CM_KEY_BODY):
    """Registry key"""

    conditions = {'os': lambda x: x == 'windows',
                  'major': lambda x: x == 6,
                  'minor': lambda x: x == 4,
                  'build': lambda x: x >= 14393,
                  }

    def full_key_name(self):
        output = []
        kcb = self.KeyControlBlock
        seen = []
        while kcb.ParentKcb and kcb.ParentKcb.obj_offset not in seen:
            if kcb.NameBlock.Name == None:
                break
            # Win10/Win2016 14393 and later skip KCB's with KEY_HIVE_ENTRY flag set
            if (kcb.Flags & KEY_FLAGS["KEY_HIVE_ENTRY"] == KEY_FLAGS["KEY_HIVE_ENTRY"]):
                kcb = kcb.ParentKcb
                if not kcb:
                    break
            output.append(str(kcb.NameBlock.Name))
            kcb = kcb.ParentKcb
            seen.append(kcb.obj_offset)
        return "\\".join(reversed(output))

class Win10Registry(obj.ProfileModification):
    """The Windows 10 registry HMAP"""

    conditions = {'os': lambda x: x == 'windows',
                  'major': lambda x: x == 6,
                  'minor': lambda x: x == 4}

    def modification(self, profile):
        profile.object_classes.update({"_HMAP_ENTRY": _HMAP_ENTRY})

class Win10x64DTB(obj.ProfileModification):
    """The Windows 10 64-bit DTB signature"""

    before = ['WindowsOverlay', 'Windows64Overlay', 'Win8x64DTB']
    conditions = {'os': lambda x: x == 'windows',
                  'major': lambda x: x == 6,
                  'minor': lambda x: x == 4,
                  'memory_model': lambda x: x == '64bit',
                  }

    def modification(self, profile):
        profile.merge_overlay({
            'VOLATILITY_MAGIC': [ None, {
            'DTBSignature' : [ None, ['VolatilityMagic', dict(value = "\x03\x00\xb6\x00")]],
            }]})

class Win10x86DTB(obj.ProfileModification):
    """The Windows 10 32-bit DTB signature"""

    before = ['WindowsOverlay', 'Win8x86DTB']
    conditions = {'os': lambda x: x == 'windows',
                  'major': lambda x: x == 6,
                  'minor': lambda x: x == 4,
                  'memory_model': lambda x: x == '32bit',
                  }

    def modification(self, profile):
        build = profile.metadata.get("build", 0)

        if build >= 15063:
            signature = "\x03\x00\x2C\x00"
        else:
            signature = "\x03\x00\x2A\x00"

        profile.merge_overlay({
            'VOLATILITY_MAGIC': [ None, {
            'DTBSignature' : [ None, ['VolatilityMagic', dict(value = signature)]],
            }]})

class Win10ObjectClasses(obj.ProfileModification):
    before = ['WindowsOverlay', 'WindowsObjectClasses']

    def modification(self, profile):
        profile.object_classes.update({'_CM_KEY_BODY' : _CM_KEY_BODY,})

class Win10KDBG(windows.AbstractKDBGMod):
    """The Windows 10 KDBG signatures"""

    before = ['Win8KDBG']
    conditions = {'os': lambda x: x == 'windows',
                  'major': lambda x: x == 6,
                  'minor': lambda x: x == 4,
                  'build': lambda x: x >= 14393}

    kdbgsize = 0x368

class ObHeaderCookieStore(object):
    """A class for finding and storing the nt!ObHeaderCookie value"""

    _instance = None

    def __init__(self):
        self._cookie = None

    def cookie(self):
        return self._cookie 

    def findcookie(self, kernel_space):
        """Find and read the nt!ObHeaderCookie value. 

        On success, return True and save the cookie value in self._cookie.
        On Failure, return False. 

        This method must be called before performing any tasks that require 
        object header validation including handles, psxview (due to pspcid) 
        and the object scanning plugins (psscan, etc). 

        NOTE: this cannot be implemented as a volatility "magic" class,
        because it must be persistent across various classes and sources. 
        We don't want to recalculate the cookie value multiple times. 
        """

        meta = kernel_space.profile.metadata 
        vers = (meta.get("major", 0), meta.get("minor", 0))

        # this algorithm only applies to Windows 10 or greater 
        if vers < (6, 4):
            return True 

        # prevent subsequent attempts from recalculating the existing value 
        if self._cookie:
            return True

        if not has_distorm:
            debug.warning("distorm3 module is not installed")
            return False 

        kdbg = tasks.get_kdbg(kernel_space)
        
        if not kdbg:
            debug.warning("Cannot find KDBG")
            return False
        
        nt_mod = None 
        
        for mod in kdbg.modules():
            nt_mod = mod 
            break 
            
        if nt_mod == None:
            debug.warning("Cannot find NT module")
            return False

        addr = nt_mod.getprocaddress("ObGetObjectType")
        if addr == None:
            debug.warning("Cannot find nt!ObGetObjectType")
            return False 

        # produce an absolute address by adding the DLL base to the RVA 
        addr += nt_mod.DllBase 
        if not nt_mod.obj_vm.is_valid_address(addr):
            debug.warning("nt!ObGetObjectType at {0} is invalid".format(addr))
            return False 

        # in theory...but so far we haven't tested 32-bits 
        model = meta.get("memory_model")    
        if model == "32bit":
            mode = distorm3.Decode32Bits
        else:
            mode = distorm3.Decode64Bits

        data = nt_mod.obj_vm.read(addr, 100)
        ops = distorm3.Decompose(addr, data, mode, distorm3.DF_STOP_ON_RET)
        addr = None

        # search backwards from the RET and find the MOVZX 

        if model == "32bit":
            # movzx ecx, byte ptr ds:_ObHeaderCookie
            for op in reversed(ops):
                if (op.size == 7 and 
                            'FLAG_DST_WR' in op.flags and
                            len(op.operands) == 2 and 
                            op.operands[0].type == 'Register' and 
                            op.operands[1].type == 'AbsoluteMemoryAddress' and 
                            op.operands[1].size == 8):
                    addr = op.operands[1].disp & 0xFFFFFFFF
                    break
        else:
            # movzx ecx, byte ptr cs:ObHeaderCookie 
            for op in reversed(ops):
                if (op.size == 7 and 
                            'FLAG_RIP_RELATIVE' in op.flags and
                            len(op.operands) == 2 and 
                            op.operands[0].type == 'Register' and 
                            op.operands[1].type == 'AbsoluteMemory' and 
                            op.operands[1].size == 8):
                    addr = op.address + op.size + op.operands[1].disp 
                    break

        if not addr:
            debug.warning("Cannot find nt!ObHeaderCookie")
            return False

        if not nt_mod.obj_vm.is_valid_address(addr):
            debug.warning("nt!ObHeaderCookie at {0} is not valid".format(addr))
            return False

        cookie = obj.Object("unsigned int", offset = addr, vm = nt_mod.obj_vm)
        self._cookie = int(cookie)

        return True

    @staticmethod
    def instance():
        if not ObHeaderCookieStore._instance:
            ObHeaderCookieStore._instance = ObHeaderCookieStore()

        return ObHeaderCookieStore._instance 

class VolatilityCookie(obj.VolatilityMagic):
    """The Windows 10 Cookie Finder"""

    def v(self):
        if self.value is None:
            return self.get_best_suggestion()
        else:
            return self.value

    def get_suggestions(self):
        if self.value:
            yield self.value
        for x in self.generate_suggestions():
            yield x

    def generate_suggestions(self):
        store = ObHeaderCookieStore.instance()
        store.findcookie(self.obj_vm)
        yield store.cookie()

class Win10Cookie(obj.ProfileModification):
    """The Windows 10 Cookie Finder"""

    before = ['WindowsOverlay']

    conditions = {'os': lambda x: x == 'windows',
                  'major': lambda x: x == 6,
                  'minor': lambda x: x == 4,
                  }

    def modification(self, profile):
        profile.merge_overlay({
            'VOLATILITY_MAGIC': [ None, {
            'ObHeaderCookie' : [ 0x0, ['VolatilityCookie', dict(configname = "COOKIE")]],
            }]})
        profile.object_classes.update({'VolatilityCookie': VolatilityCookie})

class _OBJECT_HEADER_10(win8._OBJECT_HEADER):
        
    @property
    def TypeIndex(self):
        """Wrap the TypeIndex member with a property that decodes it 
        with the nt!ObHeaderCookie value."""

        cook = obj.VolMagic(self.obj_vm).ObHeaderCookie.v()
        addr = self.obj_offset 
        indx = int(self.m("TypeIndex"))

        return ((addr >> 8) ^ cook ^ indx) & 0xFF

    def is_valid(self):
        """Determine if a given object header is valid"""

        if not obj.CType.is_valid(self):
            return False

        if self.InfoMask > 0x88:
            return False

        if self.PointerCount > 0x1000000 or self.PointerCount < 0:
            return False

        return True

    type_map = {
        2: 'Type',
        3: 'Directory',
        4: 'SymbolicLink',
        5: 'Token',
        6: 'Job',
        7: 'Process',
        8: 'Thread',
        9: 'UserApcReserve',
        10: 'IoCompletionReserve',
        11: 'Silo',
        12: 'DebugObject',
        13: 'Event',
        14: 'Mutant',
        15: 'Callback',
        16: 'Semaphore',
        17: 'Timer',
        18: 'IRTimer',
        19: 'Profile',
        20: 'KeyedEvent',
        21: 'WindowStation',
        22: 'Desktop',
        23: 'Composition',
        24: 'RawInputManager',
        25: 'TpWorkerFactory',
        26: 'Adapter',
        27: 'Controller',
        28: 'Device',
        29: 'Driver',
        30: 'IoCompletion',
        31: 'WaitCompletionPacket',
        32: 'File',
        33: 'TmTm',
        34: 'TmTx',
        35: 'TmRm',
        36: 'TmEn',
        37: 'Section',
        38: 'Session',
        39: 'Partition',
        40: 'Key',
        41: 'ALPC Port',
        42: 'PowerRequest',
        43: 'WmiGuid',
        44: 'EtwRegistration',
        45: 'EtwConsumer',
        46: 'DmaAdapter',
        47: 'DmaDomain',
        48: 'PcwObject',
        49: 'FilterConnectionPort',
        50: 'FilterCommunicationPort',
        51: 'NetworkNamespace',
        52: 'DxgkSharedResource',
        53: 'DxgkSharedSyncObject',
        54: 'DxgkSharedSwapChainObject',
        }

class _OBJECT_HEADER_10_1AC738FB(_OBJECT_HEADER_10):

    type_map = {
        2: 'Type',
        3: 'Directory',
        4: 'SymbolicLink',
        5: 'Token',
        6: 'Job',
        7: 'Process',
        8: 'Thread',
        9: 'UserApcReserve',
        10: 'IoCompletionReserve',
        11: 'DebugObject',
        12: 'Event',
        13: 'Mutant',
        14: 'Callback',
        15: 'Semaphore',
        16: 'Timer',
        17: 'IRTimer',
        18: 'Profile',
        19: 'KeyedEvent',
        20: 'WindowStation',
        21: 'Desktop',
        22: 'Composition',
        23: 'RawInputManager',
        24: 'TpWorkerFactory',
        25: 'Adapter',
        26: 'Controller',
        27: 'Device',
        28: 'Driver',
        29: 'IoCompletion',
        30: 'WaitCompletionPacket',
        31: 'File',
        32: 'TmTm',
        33: 'TmTx',
        34: 'TmRm',
        35: 'TmEn',
        36: 'Section',
        37: 'Session',
        38: 'Partition',
        39: 'Key',
        40: 'ALPC Port',
        41: 'PowerRequest',
        42: 'WmiGuid',
        43: 'EtwRegistration',
        44: 'EtwConsumer',
        45: 'DmaAdapter',
        46: 'DmaDomain',
        47: 'PcwObject',
        48: 'FilterConnectionPort',
        49: 'FilterCommunicationPort',
        50: 'NetworkNamespace',
        51: 'DxgkSharedResource',
        52: 'DxgkSharedSyncObject',
        53: 'DxgkSharedSwapChainObject',
        }

class _OBJECT_HEADER_10_DD08DD42(_OBJECT_HEADER_10):

    type_map = {
        2: 'Type',
        3: 'Directory',
        4: 'SymbolicLink',
        5: 'Token',
        6: 'Job',
        7: 'Process',
        8: 'Thread',
        9: 'UserApcReserve',
        10: 'IoCompletionReserve',
        11: 'PsSiloContextPaged',
        12: 'PsSiloContextNonPaged',
        13: 'DebugObject',
        14: 'Event',
        15: 'Mutant',
        16: 'Callback',
        17: 'Semaphore',
        18: 'Timer',
        19: 'IRTimer',
        20: 'Profile',
        21: 'KeyedEvent',
        22: 'WindowStation',
        23: 'Desktop',
        24: 'Composition',
        25: 'RawInputManager',
        26: 'CoreMessaging',
        27: 'TpWorkerFactory',
        28: 'Adapter',
        29: 'Controller',
        30: 'Device',
        31: 'Driver',
        32: 'IoCompletion',
        33: 'WaitCompletionPacket',
        34: 'File',
        35: 'TmTm',
        36: 'TmTx',
        37: 'TmRm',
        38: 'TmEn',
        39: 'Section',
        40: 'Session',
        41: 'Partition',
        42: 'Key',
        43: 'RegistryTransaction',
        44: 'ALPC',
        45: 'PowerRequest',
        46: 'WmiGuid',
        47: 'EtwRegistration',
        48: 'EtwConsumer',
        49: 'DmaAdapter',
        50: 'DmaDomain',
        51: 'PcwObject',
        52: 'FilterConnectionPort',
        53: 'FilterCommunicationPort',
        54: 'NdisCmState',
        55: 'DxgkSharedResource',
        56: 'DxgkSharedSyncObject',
        57: 'DxgkSharedSwapChainObject',
        58: 'VRegConfigurationContext',
        59: 'VirtualKey',
        }
    
class _OBJECT_HEADER_10_15063(_OBJECT_HEADER_10):

    type_map = {
		2: 'Type',
		3: 'Directory',
		4: 'SymbolicLink',
		5: 'Token',
		6: 'Job',
		7: 'Process',
		8: 'Thread',
		9: 'UserApcReserve',
		10: 'IoCompletionReserve',
		11: 'ActivityReference',
		12: 'PsSiloContextPaged',
		13: 'PsSiloContextNonPaged',
		14: 'DebugObject',
		15: 'Event',
		16: 'Mutant',
		17: 'Callback',
		18: 'Semaphore',
		19: 'Timer',
		20: 'IRTimer',
		21: 'Profile',
		22: 'KeyedEvent',
		23: 'WindowStation',
		24: 'Desktop',
		25: 'Composition',
		26: 'RawInputManager',
		27: 'CoreMessaging',
		28: 'TpWorkerFactory',
		29: 'Adapter',
		30: 'Controller',
		31: 'Device',
		32: 'Driver',
		33: 'IoCompletion',
		34: 'WaitCompletionPacket',
		35: 'File',
		36: 'TmTm',
		37: 'TmTx',
		38: 'TmRm',
		39: 'TmEn',
		40: 'Section',
		41: 'Session',
		42: 'Partition',
		43: 'Key',
		44: 'RegistryTransaction',
		45: 'ALPC Port',
		46: 'PowerRequest',
		47: 'WmiGuid',
		48: 'EtwRegistration',
		49: 'EtwSessionDemuxEntry',
		50: 'EtwConsumer',
		51: 'DmaAdapter',
		52: 'DmaDomain',
		53: 'PcwObject',
		54: 'FilterConnectionPort',
		55: 'FilterCommunicationPort',
		56: 'NdisCmState',
		57: 'DxgkSharedResource',
		58: 'DxgkSharedSyncObject',
		59: 'DxgkSharedSwapChainObject',
        60: 'DxgkCurrentDxgProcessObject',
        61: 'VRegConfigurationContext'
    	}
    
class _OBJECT_HEADER_10_16299(_OBJECT_HEADER_10):

    type_map = {
		2: 'Type',
        3: 'Directory',
        4: 'SymbolicLink',
        5: 'Token',
        6: 'Job',
        7: 'Process',
        8: 'Thread',
        9: 'Partition',
        10: 'UserApcReserve',
        11: 'IoCompletionReserve',
        12: 'ActivityReference',
        13: 'PsSiloContextPaged',
        14: 'PsSiloContextNonPaged',
        15: 'DebugObject',
        16: 'Event',
        17: 'Mutant',
        18: 'Callback',
        19: 'Semaphore',
        20: 'Timer',
        21: 'IRTimer',
        22: 'Profile',
        23: 'KeyedEvent',
        24: 'WindowStation',
        25: 'Desktop',
        26: 'Composition',
        27: 'RawInputManager',
        28: 'CoreMessaging',
        29: 'TpWorkerFactory',
        30: 'Adapter',
        31: 'Controller',
        32: 'Device',
        33: 'Driver',
        34: 'IoCompletion',
        35: 'WaitCompletionPacket',
        36: 'File',
        37: 'TmTm',
        38: 'TmTx',
        39: 'TmRm',
        40: 'TmEn',
        41: 'Section',
        42: 'Session',
        43: 'Key',
        44: 'RegistryTransaction',
        45: 'ALPC Port',
        46: 'EnergyTracker',
        47: 'PowerRequest',
        48: 'WmiGuid',
        49: 'EtwRegistration',
        50: 'EtwSessionDemuxEntry',
        51: 'EtwConsumer',
        52: 'DmaAdapter',
        53: 'DmaDomain',
        54: 'PcwObject',
        55: 'FilterConnectionPort',
        56: 'FilterCommunicationPort',
        57: 'NdisCmState',
        58: 'DxgkSharedResource',
        59: 'DxgkSharedSyncObject',
        60: 'DxgkSharedSwapChainObject',
        61: 'DxgkDisplayManagerObject',
        62: 'DxgkCurrentDxgProcessObject',
        63: 'DxgkSharedProtectedSessionObject',
        64: 'DxgkSharedBundleObject',
        65: 'VRegConfigurationContext',
    	}

class _OBJECT_HEADER_10_17134(_OBJECT_HEADER_10):

    type_map = {
        2: "Type",
        3: "Directory",
        4: "SymbolicLink",
        5: "Token",
        6: "Job",
        7: "Process",
        8: "Thread",
        9: "Partition",
        10: "UserApcReserve",
        11: "IoCompletionReserve",
        12: "ActivityReference",
        13: "PsSiloContextPaged",
        14: "PsSiloContextNonPaged",
        15: "DebugObject",
        16: "Event",
        17: "Mutant",
        18: "Callback",
        19: "Semaphore",
        20: "Timer",
        21: "IRTimer",
        22: "Profile",
        23: "KeyedEvent",
        24: "WindowStation",
        25: "Desktop",
        26: "Composition",
        27: "RawInputManager",
        28: "CoreMessaging",
        29: "TpWorkerFactory",
        30: "Adapter",
        31: "Controller",
        32: "Device",
        33: "Driver",
        34: "IoCompletion",
        35: "WaitCompletionPacket",
        36: "File",
        37: "TmTm",
        38: "TmTx",
        39: "TmRm",
        40: "TmEn",
        41: "Section",
        42: "Session",
        43: "Key",
        44: "RegistryTransaction",
        45: "ALPC",
        46: "EnergyTracker",
        47: "PowerRequest",
        48: "WmiGuid",
        49: "EtwRegistration",
        50: "EtwSessionDemuxEntry",
        51: "EtwConsumer",
        52: "CoverageSampler",
        53: "DmaAdapter",
        54: "PcwObject",
        55: "FilterConnectionPort",
        56: "FilterCommunicationPort",
        57: "NdisCmState",
        58: "DxgkSharedResource",
        59: "DxgkSharedKeyedMutexObject",
        60: "DxgkSharedSyncObject",
        61: "DxgkSharedSwapChainObject",
        62: "DxgkDisplayManagerObject",
        63: "DxgkCurrentDxgProcessObject",
        64: "DxgkSharedProtectedSessionObject",
        65: "DxgkSharedBundleObject",
        66: "DxgkCompositionObject",
        67: "VRegConfigurationContext",
    }

class _HANDLE_TABLE_10_DD08DD42(win8._HANDLE_TABLE_81R264):
    
    def decode_pointer(self, value):
        
        value = value & 0xFFFFFFFFFFFFFFF8
        value = value >> self.DECODE_MAGIC
        if (value & (1 << 47)):
            value = value | 0xFFFF000000000000
    
        return value

class Win10ObjectHeader(obj.ProfileModification):
    before = ["Win8ObjectClasses"]
    conditions = {'os': lambda x: x == 'windows',
                  'major': lambda x: x == 6,
                  'minor': lambda x: x == 4}

    def modification(self, profile):

        metadata = profile.metadata
        build = metadata.get("build", 0)

        if build >= 17134:
            header = _OBJECT_HEADER_10_17134

            ## update the handle table here as well
            if metadata.get("memory_model") == "64bit":
                profile.object_classes.update({
                    "_HANDLE_TABLE": _HANDLE_TABLE_10_DD08DD42})

        elif build >= 16299:
            header = _OBJECT_HEADER_10_16299

            ## update the handle table here as well
            if metadata.get("memory_model") == "64bit":
                profile.object_classes.update({
                    "_HANDLE_TABLE": _HANDLE_TABLE_10_DD08DD42})

        elif build >= 15063:
            header = _OBJECT_HEADER_10_15063

            ## update the handle table here as well
            if metadata.get("memory_model") == "64bit":
                profile.object_classes.update({
                    "_HANDLE_TABLE": _HANDLE_TABLE_10_DD08DD42})

        elif build >= 14393:
            header = _OBJECT_HEADER_10_DD08DD42
            
            ## update the handle table here as well
            if metadata.get("memory_model") == "64bit":
                profile.object_classes.update({
                    "_HANDLE_TABLE": _HANDLE_TABLE_10_DD08DD42})
            
        elif build >= 10586:
            header = _OBJECT_HEADER_10_1AC738FB
        else:
            header = _OBJECT_HEADER_10

        profile.object_classes.update({"_OBJECT_HEADER": header})

class WSLPicoModifcation(obj.ProfileModification): 
    """Profile modification for Windows Subsystem for Linux, 
    in particular the Pico process contexts"""

    before = ['WindowsOverlay']
    conditions = {'os': lambda x: x == 'windows',
                  'major': lambda x: x == 6,
                  'minor': lambda x: x == 4, 
                  'memory_model': lambda x: x == '64bit'}
                  
    def modification(self, profile):
        
        build = profile.metadata.get("build", 0)
        
        if build <= 14393:
            # offsets for anniversary update
            pico_context = {'_PICO_CONTEXT' : [ None, {
                "Name": [ 0x178, ["_UNICODE_STRING"]]}]}
        else:
            # offsets for creators & fall creators  
            pico_context = {'_PICO_CONTEXT' : [ None, {
                "Name": [ 0x180, ["_UNICODE_STRING"]]}]}
        
        profile.vtypes.update(pico_context)
        
        profile.merge_overlay({'_EPROCESS': [ None, {
            'PicoContext' : [ None, ['pointer', ['_PICO_CONTEXT']]],
            }]})

class Win10PoolHeader(obj.ProfileModification):
    before = ['WindowsOverlay']
    conditions = {'os': lambda x: x == 'windows',
                  'major': lambda x: x == 6,
                  'minor': lambda x: x == 4,
                  'build': lambda x: x == 10586}

    def modification(self, profile):

        meta = profile.metadata
        memory_model = meta.get("memory_model", "32bit")

        if memory_model == "32bit":
            pool_types = {'_POOL_HEADER' : [ 0x8, {
                'PreviousSize' : [ 0x0, ['BitField', dict(start_bit = 0, end_bit = 9, native_type='unsigned short')]],
                'PoolIndex' : [ 0x0, ['BitField', dict(start_bit = 9, end_bit = 16, native_type='unsigned short')]],
                'BlockSize' : [ 0x2, ['BitField', dict(start_bit = 0, end_bit = 9, native_type='unsigned short')]],
                'PoolType' : [ 0x2, ['BitField', dict(start_bit = 9, end_bit = 16, native_type='unsigned short')]],
                'Ulong1' : [ 0x0, ['unsigned long']],
                'PoolTag' : [ 0x4, ['unsigned long']],
                'AllocatorBackTraceIndex' : [ 0x4, ['unsigned short']],
                'PoolTagHash' : [ 0x6, ['unsigned short']],
                }]}
        else:
            pool_types = {'_POOL_HEADER' : [ 0x10, {
                 'PreviousSize' : [ 0x0, ['BitField', dict(start_bit = 0, end_bit = 8, native_type='unsigned short')]],
                 'PoolIndex' : [ 0x0, ['BitField', dict(start_bit = 8, end_bit = 16, native_type='unsigned short')]],
                 'BlockSize' : [ 0x2, ['BitField', dict(start_bit = 0, end_bit = 8, native_type='unsigned short')]],
                 'PoolType' : [ 0x2, ['BitField', dict(start_bit = 8, end_bit = 16, native_type='unsigned short')]],
                 'Ulong1' : [ 0x0, ['unsigned long']],
                 'PoolTag' : [ 0x4, ['unsigned long']],
                 'ProcessBilled' : [ 0x8, ['pointer64', ['_EPROCESS']]],
                 'AllocatorBackTraceIndex' : [ 0x8, ['unsigned short']],
                 'PoolTagHash' : [ 0xa, ['unsigned short']],
                 }]}

        profile.vtypes.update(pool_types)

class Win10x64(obj.Profile):
    """ A Profile for Windows 10 x64 """
    _md_memory_model = '64bit'
    _md_os = 'windows'
    _md_major = 6
    _md_minor = 4
    _md_build = 9841
    _md_vtype_module = 'volatility.plugins.overlays.windows.win10_x64_vtypes'
    _md_product = ["NtProductWinNt"]

class Win10x64_10240_17770(obj.Profile):
    """ A Profile for Windows 10 x64 (10.0.10240.17770 / 2018-02-10) """
    _md_memory_model = '64bit'
    _md_os = 'windows'
    _md_major = 6
    _md_minor = 4
    _md_build = 10240
    _md_vtype_module = 'volatility.plugins.overlays.windows.win10_x64_10240_17770_vtypes'
    _md_product = ["NtProductWinNt"]

class Win10x64_10586(obj.Profile):
    """ A Profile for Windows 10 x64 (10.0.10586.306 / 2016-04-23) """
    _md_memory_model = '64bit'
    _md_os = 'windows'
    _md_major = 6
    _md_minor = 4
    _md_build = 10586
    _md_vtype_module = 'volatility.plugins.overlays.windows.win10_x64_1AC738FB_vtypes'
    _md_product = ["NtProductWinNt"]

class Win10x64_14393(obj.Profile):
    """ A Profile for Windows 10 x64 (10.0.14393.0 / 2016-07-16) """
    _md_memory_model = '64bit'
    _md_os = 'windows'
    _md_major = 6
    _md_minor = 4
    _md_build = 14393
    _md_vtype_module = 'volatility.plugins.overlays.windows.win10_x64_DD08DD42_vtypes'
    _md_product = ["NtProductWinNt"]

class Win10x86(obj.Profile):
    """ A Profile for Windows 10 x86 """
    _md_memory_model = '32bit'
    _md_os = 'windows'
    _md_major = 6
    _md_minor = 4
    _md_build = 9841
    _md_vtype_module = 'volatility.plugins.overlays.windows.win10_x86_vtypes'
    _md_product = ["NtProductWinNt"]

class Win10x86_10240_17770(obj.Profile):
    """ A Profile for Windows 10 x86 (10.0.10240.17770 / 2018-02-10) """
    _md_memory_model = '32bit'
    _md_os = 'windows'
    _md_major = 6
    _md_minor = 4
    _md_build = 10240
    _md_vtype_module = 'volatility.plugins.overlays.windows.win10_x86_10240_17770_vtypes'
    _md_product = ["NtProductWinNt"]

class Win10x86_10586(obj.Profile):
    """ A Profile for Windows 10 x86 (10.0.10586.420 / 2016-05-28) """
    _md_memory_model = '32bit'
    _md_os = 'windows'
    _md_major = 6
    _md_minor = 4
    _md_build = 10586
    _md_vtype_module = 'volatility.plugins.overlays.windows.win10_x86_44B89EEA_vtypes'
    _md_product = ["NtProductWinNt"]

class Win10x86_14393(obj.Profile):
    """ A Profile for Windows 10 x86 (10.0.14393.0 / 2016-07-16) """
    _md_memory_model = '32bit'
    _md_os = 'windows'
    _md_major = 6
    _md_minor = 4
    _md_build = 14393
    _md_vtype_module = 'volatility.plugins.overlays.windows.win10_x86_9619274A_vtypes'
    _md_product = ["NtProductWinNt"]
    
class Win2016x64_14393(Win10x64_14393):
    """ A Profile for Windows Server 2016 x64 (10.0.14393.0 / 2016-07-16) """
    _md_memory_model = '64bit'
    _md_os = 'windows'
    _md_major = 6
    _md_minor = 4
    _md_build = 14393
    _md_vtype_module = 'volatility.plugins.overlays.windows.win10_x64_DD08DD42_vtypes'
    _md_product = ["NtProductLanManNt", "NtProductServer"]

class Win10x86_15063(obj.Profile):
    """ A Profile for Windows 10 x86 (10.0.15063.0 / 2017-04-04) """
    _md_memory_model = '32bit'
    _md_os = 'windows'
    _md_major = 6
    _md_minor = 4
    _md_build = 15063
    _md_vtype_module = 'volatility.plugins.overlays.windows.win10_x86_15063_vtypes'
    _md_product = ["NtProductWinNt"]

class Win10x86_16299(obj.Profile):
    """ A Profile for Windows 10 x86 (10.0.16299.15 / 2017-09-29) """
    _md_memory_model = '32bit'
    _md_os = 'windows'
    _md_major = 6
    _md_minor = 4
    _md_build = 16299
    _md_vtype_module = 'volatility.plugins.overlays.windows.win10_x86_16299_vtypes'
    _md_product = ["NtProductWinNt"]

class Win10x86_17134(obj.Profile):
    """ A Profile for Windows 10 x86 (10.0.17134.1 / 2018-04-11) """
    _md_memory_model = '32bit'
    _md_os = 'windows'
    _md_major = 6
    _md_minor = 4
    _md_build = 17134
    _md_vtype_module = 'volatility.plugins.overlays.windows.win10_x86_17134_vtypes'
    _md_product = ["NtProductWinNt"]

class Win10x64_15063(obj.Profile):
    """ A Profile for Windows 10 x64 (10.0.15063.0 / 2017-04-04) """
    _md_memory_model = '64bit'
    _md_os = 'windows'
    _md_major = 6
    _md_minor = 4
    _md_build = 15063
    _md_vtype_module = 'volatility.plugins.overlays.windows.win10_x64_15063_vtypes'
    _md_product = ["NtProductWinNt"]
    
class Win10x64_16299(obj.Profile):
    """ A Profile for Windows 10 x64 (10.0.16299.0 / 2017-09-22) """
    _md_memory_model = '64bit'
    _md_os = 'windows'
    _md_major = 6
    _md_minor = 4
    _md_build = 16299
    _md_vtype_module = 'volatility.plugins.overlays.windows.win10_x64_16299_vtypes'
    _md_product = ["NtProductWinNt"]

class Win10x64_17134(obj.Profile):
    """ A Profile for Windows 10 x64 (10.0.17134.1 / 2018-04-11) """
    _md_memory_model = '64bit'
    _md_os = 'windows'
    _md_major = 6
    _md_minor = 4
    _md_build = 17134
    _md_vtype_module = 'volatility.plugins.overlays.windows.win10_x64_17134_vtypes'
    _md_product = ["NtProductWinNt"]


# Volatility
# Copyright (C) 2007-2013 Volatility Foundation
#
# This file is part of Volatility.
#
# Volatility is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# Volatility is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Volatility.  If not, see <http://www.gnu.org/licenses/>.
#

"""
@author:       Andrew Case
@license:      GNU General Public License 2.0
@contact:      atcuno@gmail.com
@organization: 
"""

import os
import volatility.plugins.mac.common as common
import volatility.plugins.mac.mount as mac_mount
import volatility.obj as obj

class mac_list_files(common.AbstractMacCommand):
    """ Lists files in the file cache """

    def __init__(self, config, *args, **kwargs):
        common.AbstractMacCommand.__init__(self, config, *args, **kwargs)
        
        self._config.add_option('SHOW_ORPHANS', 
            short_option = 's', 
            default = False, 
            help = 'Show orphans (vnodes without a parent)', 
            action = 'store_true')

    @staticmethod
    def list_files(config):
    
        plugin = mac_mount.mac_mount(config)
        mounts = plugin.calculate()
        vnodes = {}
        parent_vnodes = {}

        ## build an initial table of all vnodes 
        for mount in mounts:
            vnode = mount.mnt_vnodelist.tqh_first.dereference()

            while vnode:
                ## abort here to prevent going in a loop 
                if vnode.obj_offset in vnodes:
                    break
                 
                ## its ok to call the slower full_path() 
                ## here because its only done for root 
                ## nodes which is only a couple per system
                if int(vnode.v_flag) & 1:
                    name  = vnode.full_path()
                
                    entry = [name, None, vnode]
                    vnodes[vnode.obj_offset] = entry
 
                else:
                    name = vnode.v_name.dereference()
                    parent = vnode.v_parent.dereference()
                
                    if parent:
                        par_offset = parent.obj_offset 
                    else: 
                        if config.SHOW_ORPHANS:
                            par_offset = None
                        else:
                            vnode = vnode.v_mntvnodes.tqe_next.dereference() 
                            continue
            
                    entry = [name, par_offset, vnode]
                    vnodes[vnode.obj_offset] = entry
                
                vnode = vnode.v_mntvnodes.tqe_next.dereference() 

        ## account for vnodes that aren't in the list but are 
        ## referenced from other vnode's v_parent pointers 
        for key, val in vnodes.items():
            name, parent, vnode = val    
            
            if not name or not parent:
                continue
                
            parent = obj.Object("vnode", 
                offset = parent, 
                vm = vnode.obj_vm)
                
            while parent:
            
                if parent.obj_offset in vnodes:
                    break

                name = parent.v_name.dereference()
                next_parent = parent.v_parent.dereference()
            
                if next_parent:
                    par_offset = next_parent.obj_offset 
                else: 
                    par_offset = None
        
                entry = [str(name), par_offset, parent]
                vnodes[parent.obj_offset] = entry
                
                parent = next_parent  

        ## build the full paths for all directories
        for key, val in vnodes.items():
            name, parent, vnode = val

            ## we can't have unnamed files or directories
            if not name:
                continue
    
            if not vnode.is_dir():
                continue
  
            name = str(name)
            
            if parent in parent_vnodes:
                full_path = parent_vnodes[parent] + "/" + name
            else:
                paths = [name]
                while parent:
                    entry = vnodes.get(parent)
                
                    ## a vnode's parent wasn't found or 
                    ## we reached the root directory 
                    if not entry:
                        break
                    
                    name, parent, _vnode = entry
                    if not name:
                        break
                    
                    paths.append(str(name))
                
                ## build the path in reverse order 
                full_path = "/".join(reversed(paths))
                
            parent_vnodes[key] = full_path

        ## link everything up with their parents 
        for val in vnodes.values():
            name, parent, vnode = val
            
            if not name:
                continue
           
            name = str(name)
 
            entry = parent_vnodes.get(parent) 
            if not entry:
                yield vnode, name
            else:
                full_path = entry + "/" + name
                
                ## add a leading slash if one doesn't exist
                if full_path[0] != "/":
                    full_path = "/" + full_path
                    
                ## otherwise in some cases we may have double 
                ## slashes so reduce that down to just one 
                elif full_path[0:2] == "//":
                    full_path = full_path[1:]

                yield vnode, full_path

    def calculate(self):
        common.set_plugin_members(self)
        config = self._config
        
        for result in mac_list_files.list_files(config):
            yield result

    def render_text(self, outfd, data):
        self.table_header(outfd, [("Offset (V)", "[addrpad]"), ("File Path", "")])
        for vnode, path in data:
            self.table_row(outfd, vnode.obj_offset, path)    



# Volatility
# Copyright (C) 2007-2013 Volatility Foundation
#
# This file is part of Volatility.
#
# Volatility is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# Volatility is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Volatility.  If not, see <http://www.gnu.org/licenses/>.
#

"""
@author:       Andrew Case
@license:      GNU General Public License 2.0
@contact:      atcuno@gmail.com
@organization: 
"""
import volatility.obj as obj
import volatility.plugins.mac.common as common
import volatility.plugins.mac.pstasks as pstasks

class mac_kevents(common.AbstractMacCommand):
    """ Show parent/child relationship of processes """

    def _walk_karray(self, address, count):
        arr = obj.Object(theType = "Array", targetType = "klist", offset = address, vm = self.addr_space, count = count) 

        for klist in arr:
                kn = klist.slh_first

                while kn.is_valid():
                    yield kn
 
                    kn = kn.kn_link.sle_next
 
    def calculate(self):
        common.set_plugin_members(self)
   
        for task in pstasks.mac_tasks(self._config).calculate():
            fdp = task.p_fd
    
            # for (i = 0; i < fdp->fd_knlistsize; i++) {
            #    kn = SLIST_FIRST(&fdp->fd_knlist[i]);
            for kn in self._walk_karray(fdp.fd_knlist, fdp.fd_knlistsize):
                yield task, kn
            
            # if (fdp->fd_knhashmask != 0) {
            #    for (i = 0; i < (int)fdp->fd_knhashmask + 1; i++) {
            #        kn = SLIST_FIRST(&fdp->fd_knhash[i]);
            mask = fdp.fd_knhashmask             
            if mask != 0:
                for kn in self._walk_karray(fdp.fd_knhash, mask + 1):
                    yield task, kn


            kn = task.p_klist.slh_first
            while kn.is_valid():
                
                yield task, kn

                kn = kn.kn_link.sle_next

    def _get_flags(self, fflags, filters): 
        context = ""

        if fflags != 0:
            for (flag, idx) in filters:
                if fflags & idx == idx:
                    context = context + flag + ", "   

            if len(context) > 2 and context[-2:] == ", ":
                context = context[:-2]                     

        return context

    def render_text(self, outfd, data):
        event_types = ["INVALID EVENT", "EVFILT_READ", "EVFILT_WRITE", "EVFILT_AIO", "EVFILT_VNODE", "EVFILT_PROC", "EVFILT_SIGNAL"]
        event_types = event_types + ["EVFILT_TIMER", "EVFILT_MACHPORT", "EVFILT_FS", "EVFILT_USER", "INVALID EVENT", "EVFILT_VM"]
                
        vnode_filt = [("NOTE_DELETE", 1), ("NOTE_WRITE", 2), ("NOTE_EXTEND", 4), ("NOTE_ATTRIB", 8)]
        vnode_filt = vnode_filt + [("NOTE_LINK", 0x10), ("NOTE_RENAME", 0x20), ("NOTE_REVOKE", 0x40)]          
 
        proc_filt  = [("NOTE_EXIT", 0x80000000), ("NOTE_EXITSTATUS", 0x04000000), ("NOTE_FORK", 0x40000000)]
        proc_filt  = proc_filt + [("NOTE_EXEC", 0x20000000), ("NOTE_SIGNAL", 0x08000000), ("NOTE_REAP", 0x10000000)]

        time_filt = [("NOTE_SECONDS", 1), ("NOTE_USECONDS", 2), ("NOTE_NSECONDS", 4), ("NOTE_ABSOLUTE", 8)]

        self.table_header(outfd, [("Offset", "[addrpad]"),
                          ("Name", "20"),
                          ("Pid", "8"),
                          ("Ident", "6"),
                          ("Filter", "20"),
                          ("Context", ""),])

        for task, kn in data:
            filt_idx = kn.kn_kevent.filter * -1
            if 0 < filt_idx < len(event_types):
                fname = event_types[filt_idx]
            else:
                continue

            context = ""    
            fflags  = kn.kn_sfflags

            # EVFILT_VNODE
            if filt_idx == 4:
                context = self._get_flags(fflags, vnode_filt)
 
            # EVFILT_PROC
            elif filt_idx == 5:
                context = self._get_flags(fflags, proc_filt) 

            elif filt_idx == 7:
                context = self._get_flags(fflags, time_filt) 

            self.table_row(outfd, kn.v(), str(task.p_comm), task.p_pid, kn.kn_kevent.ident, fname, context)   










# Volatility
# Copyright (c) 2008-2013 Volatility Foundation
#
# This file is part of Volatility.
#
# Volatility is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License Version 2 as
# published by the Free Software Foundation.  You may not use, modify or
# distribute this program under any other version of the GNU General
# Public License.
#
# Volatility is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Volatility.  If not, see <http://www.gnu.org/licenses/>.
#

import struct
import volatility.obj as obj
import volatility.addrspace as addrspace
import volatility.constants as constants
import volatility.utils as utils
import volatility.plugins.overlays.windows.win8 as win8
import volatility.plugins.overlays.windows.windows as windows_main
import volatility.plugins.patchguard as patchguard
import volatility.registry as registry

try:
    import distorm3
    has_distorm = True
except ImportError:
    has_distorm = False

class VolatilityKDBG(obj.VolatilityMagic):
    """A Scanner for KDBG data within an address space. 

    This implementation is specific for Windows 8 / 2012 
    64-bit versions because the KDBG block is encoded. We 
    have to find it a special way and then perform the 
    decoding routine before Volatility plugins can run. 
    """

    def v(self):
        """The --kdbg parameter for this Win8/2012 KDBG 
        structure is the virtual address of the 
        nt!KdCopyDataBlock function (see kdbgscan output).
        """

        if self.value is None:
            return self.get_best_suggestion()
        else:
            return self.copy_data_block(self.value)

    def get_suggestions(self):
        if self.value:
            yield self.copy_data_block(self.value)
        for x in self.generate_suggestions():
            yield x

    def decode_kdbg(self, vals):
        """Decoder the KDBG block using the provided 
        magic values and the algorithm reversed from 
        the Windows kernel file."""

        block_encoded, kdbg_block, wait_never, wait_always = vals
        # just take the maximum. if we decode a tiny bit of 
        # extra data in some cases, its totally fine.
        kdbg_size = max(self.unique_sizes())
        buffer = ""

        entries = obj.Object("Array", 
                            targetType = "unsigned long long", 
                            count = kdbg_size / 8, 
                            offset = kdbg_block, vm = self.obj_vm)

        for entry in entries: 
            low_byte = (wait_never & 0xFFFFFFFF) & 0xFF
            entry = patchguard.rol(entry ^ wait_never, low_byte)
            swap_xor = block_encoded.obj_offset | 0xFFFF000000000000
            entry = patchguard.bswap(entry ^ swap_xor)
            buffer += struct.pack("Q", entry ^ wait_always) 

        return buffer

    def unique_sizes(self):
    	"""Determine the possible KDBG sizes to scan for, across all 
    	profiles Win8 x64 and above. We do this by reflecting back on 
    	the profile modifications to see which ones would trigger and
    	then grabbing the KDBG size."""
    
        items = registry.get_plugin_classes(windows_main.AbstractKDBGMod).items()
        sizes = set()
        
        for name, cls in items:
            try:
                if (not cls.conditions["os"]("windows") or 
                        not cls.conditions["major"](6)):
                    continue

                sizes.add(cls.kdbgsize)
            except:
                continue

        return sizes

    def copy_data_block(self, full_addr):
        """This function emulates nt!KdCopyDataBlock on a live 
        machine by finding the encoded KDBG structure and using
        the required entropy values to decode it."""

        sizes = self.unique_sizes()
        alignment = 8 
        addr_space = self.obj_vm
        bits = distorm3.Decode64Bits

        # nt!KdCopyDataBlock is about 100 bytes, we don't want to read
        # too little and truncate the function, but too much will reach
        # into other function's space
        code = addr_space.read(full_addr, 300)

        # potentially we crossed a boundary into swapped or unallocated space
        if code == None:
            return obj.NoneObject("Crossed a code boundary")

        found_size = False 
        
        for size in sizes:
            val = struct.pack("I", size / alignment)
            if code.find(val) != -1:
                found_size = True
                break
        
        if not found_size:
            return obj.NoneObject("Cannot find KDBG size signature")

        version = (addr_space.profile.metadata.get('major', 0), addr_space.profile.metadata.get('minor', 0))
        if version < (6, 4):
            # we don't perform this check for Windows 10.x
            found_str = False 
            
            for size in sizes:
                val = struct.pack("I", size)
                if code.find(val) != -1:
                    found_str = True
                    break
                
            if not found_str:
                return obj.NoneObject("Cannot find KDBG size signature")  

        ops = list(distorm3.Decompose(full_addr, code, bits))

        # nt!KdDebuggerDataBlock
        kdbg_block = None
        # nt!KiWaitNever
        wait_never = None
        # nt!KiWaitAlways
        wait_always = None
        # nt!KdpDataBlockEncoded
        block_encoded = None
        
        # collect instructions up to the first RET
        before_ret = []
        # we need a bswap instruction to be valid
        found_bswap = False
        
        for op in ops:
        	if op.mnemonic == "BSWAP":
        		found_bswap = True
        	elif op.mnemonic == "RET":
        		break
        	else:
        		before_ret.append(op)
        		
        if not found_bswap:
        	return obj.NoneObject("No bswap instruction found")

        for op in before_ret:
            # cmp cs:KdpDataBlockEncoded, 0
            if (not block_encoded and op.mnemonic == "CMP" and 
                        op.operands[0].type == "AbsoluteMemory" and 
                        op.operands[1].type == "Immediate" and 
                        op.operands[1].value == 0):
                # an x64 RIP turned absolute 
                offset = op.address + op.size + op.operands[0].disp
                block_encoded = obj.Object("unsigned char", 
                                        offset = offset,
                                        vm = addr_space)
            # lea rdx, KdDebuggerDataBlock
            elif (not kdbg_block and op.mnemonic == "LEA" and 
                        op.operands[0].type == "Register" and 
                        op.operands[0].size == 64 and 
                        op.operands[1].type == "AbsoluteMemory" and 
                        op.operands[1].dispSize == 32):
                kdbg_block = op.address + op.size + op.operands[1].disp 
            # mov r10, cs:KiWaitNever
            elif (not wait_never and op.mnemonic == "MOV" and 
                        op.operands[0].type == "Register" and 
                        op.operands[0].size == 64 and 
                        op.operands[1].type == "AbsoluteMemory" and 
                        op.operands[1].dispSize == 32):
                offset = op.address + op.size + op.operands[1].disp
                wait_never = obj.Object("unsigned long long", 
                                        offset = offset, 
                                        vm = addr_space)
            # mov r11, cs:KiWaitAlways (Win 8 x64)
            # xor rdx, cs:KiWaitAlways (Win 8.1 x64)
            elif (not wait_always and op.mnemonic in ["MOV", "XOR"] and 
                        op.operands[0].type == "Register" and 
                        op.operands[0].size == 64 and 
                        op.operands[1].type == "AbsoluteMemory" and 
                        op.operands[1].dispSize == 32):
                offset = op.address + op.size + op.operands[1].disp
                wait_always = obj.Object("unsigned long long", 
                                        offset = offset,
                                        vm = addr_space)
                break

        # check if we've found all the required offsets 
        if (block_encoded != None 
                    and kdbg_block != None 
                    and wait_never != None
                    and wait_always != None):
            
            # some acquisition tools decode the KDBG block but leave 
            # nt!KdpDataBlockEncoded set, so we handle it here. 
            tag_offset = addr_space.profile.get_obj_offset("_DBGKD_DEBUG_DATA_HEADER64", "OwnerTag")
            signature = addr_space.read(kdbg_block + tag_offset, 4)

            if block_encoded == 1 and signature != "KDBG":
                vals = block_encoded, kdbg_block, wait_never, wait_always
                data = self.decode_kdbg(vals)
                buff = addrspace.BufferAddressSpace(
                            config = addr_space.get_config(),
                            base_offset = kdbg_block,
                            data = data)
                kdbg = obj.Object("_KDDEBUGGER_DATA64", 
                            offset = kdbg_block, 
                            vm = buff, 
                            native_vm = addr_space)
            else:
                kdbg = obj.Object("_KDDEBUGGER_DATA64", 
                            offset = kdbg_block, 
                            vm = addr_space)

            kdbg.newattr('KdCopyDataBlock', full_addr)
            kdbg.newattr('block_encoded', block_encoded == 1 and signature != "KDBG")
            kdbg.newattr('wait_never', wait_never)
            kdbg.newattr('wait_always', wait_always)                    

            if kdbg.Header.OwnerTag == 0x4742444b:
                return kdbg
                
        return obj.NoneObject("Cannot find decoding entropy values")

    def generate_suggestions(self):
        """Generates a list of possible KDBG structure locations"""

        if not has_distorm:
            raise StopIteration("The distorm3 Python library is required")

        overlap = 20
        offset = 0 
        current_offset = offset
        addr_space = self.obj_vm
        
        addresses = sorted(addr_space.get_available_addresses())
        for (range_start, range_size) in addresses:
            # Jump to the next available point to scan from
            current_offset = max(range_start, current_offset)
            range_end = range_start + range_size

            if current_offset < 0xf80000000000:
                continue

            while (current_offset < range_end):
                # Figure out how much data to read
                l = min(constants.SCAN_BLOCKSIZE + overlap, range_end - current_offset)

                data = addr_space.zread(current_offset, l)
            
                for addr in utils.iterfind(data, "\x80\x3D"):
                    full_addr = addr + current_offset 
                    result = self.copy_data_block(full_addr)
                    if result:
                        yield result

                current_offset += min(constants.SCAN_BLOCKSIZE, l)

class Win8x64VolatilityKDBG(obj.ProfileModification):
    """Apply the KDBG finder for x64"""

    before = ['WindowsOverlay', 'WindowsObjectClasses']
    conditions = {'os': lambda x: x == 'windows',
                  'major': lambda x: x == 6,
                  'minor': lambda x: x >= 2, 
                  'memory_model': lambda x: x == "64bit"}
    
    def modification(self, profile):
        profile.object_classes.update({"VolatilityKDBG": VolatilityKDBG})



import os
import sys

datas = []

for path in sys.path:
    if os.path.exists(os.path.join(path, "yara.pyd")):
        datas.append((os.path.join(path, "yara.pyd"), "."))
    if os.path.exists(os.path.join(path, "yara.so")):
        datas.append((os.path.join(path, "yara.so"), "."))



# Volatility
# Copyright (C) 2007-2017 Volatility Foundation
# Copyright (C) 2017 Michael Hale Ligh <michael.ligh@mnin.org>
#
# This file is part of Volatility.
#
# Volatility is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# Volatility is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Volatility.  If not, see <http://www.gnu.org/licenses/>.
#

import volatility.obj as obj 

class Win10x86_Gui(obj.ProfileModification):

    before = ["XP2003x86BaseVTypes", "Win32Kx86VTypes", "AtomTablex86Overlay", "Win32KCoreClasses", "Win8x86Gui"]

    conditions = {'os': lambda x: x == 'windows',
                  'memory_model': lambda x: x == '32bit',
                  'major': lambda x: x == 6,
                  'minor': lambda x: x == 4}

    def modification(self, profile):
        build = profile.metadata.get('build', 0)
    
        if build >= 15063:    
            profile.merge_overlay({
                'tagDESKTOP': [None, {
                    'rpdeskNext': [0x10, ['pointer', ['tagDESKTOP']]],
                    'rpwinstaParent': [0x14, ['pointer', ['tagWINDOWSTATION']]],
                    'pheapDesktop': [0x40, ['pointer', ['tagWIN32HEAP']]],
                    'PtiList': [0x5c, ['_LIST_ENTRY']],
                }],
                'tagTHREADINFO': [None, {
                    'ppi': [0xe0, ['pointer', ['tagPROCESSINFO']]],
                    'PtiLink': [0x188, ['_LIST_ENTRY']],
                }],
                'tagWND': [None, {
                    'spwndNext': [0x34, ['pointer', ['tagWND']]],
                    'spwndPrev': [0x38, ['pointer', ['tagWND']]],
                    'spwndParent': [0x3c, ['pointer', ['tagWND']]],
                    'spwndChild': [0x40, ['pointer', ['tagWND']]],
                    'lpfnWndProc': [0x68, ['pointer', ['void']]], #?
                    'pcls': [0x6c, ['pointer', ['tagCLS']]], #?
                    'strName': [0x8c, ['_LARGE_UNICODE_STRING']], #?
                }],
            })

class Win10x64_Gui(obj.ProfileModification):

    before = ["Win32KCoreClasses", "Win8x64Gui"]

    conditions = {'os': lambda x: x == 'windows',
                  'memory_model': lambda x: x == '64bit',
                  'major': lambda x: x == 6,
                  'minor': lambda x: x == 4}

    def modification(self, profile):
        build = profile.metadata.get('build', 0)
    
        if build >= 15063:    
            profile.merge_overlay({
                'tagDESKTOP': [None, {
                    'rpdeskNext': [0x20, ['pointer64', ['tagDESKTOP']]],
                    'rpwinstaParent': [0x28, ['pointer64', ['tagWINDOWSTATION']]],
                    'pheapDesktop': [0x80, ['pointer', ['tagWIN32HEAP']]],
                    'PtiList': [0xA8, ['_LIST_ENTRY']],
                }],
                'tagTHREADINFO': [None, {
                    'ppi': [0x190, ['pointer', ['tagPROCESSINFO']]],
                     # zzzSetDesktop 
                     # mov [rbx+1B8h], rax
                     # lea rax, [rbx+2C8h]
                     # lea rcx, [rdi+0A8h] => int 29h 
                    'PtiLink': [0x2c8, ['_LIST_ENTRY']],
                }],
                'tagWND': [None, {
                    'spwndNext': [0x58, ['pointer64', ['tagWND']]],
                    'spwndPrev': [0x60, ['pointer64', ['tagWND']]],
                    'spwndParent': [0x68, ['pointer64', ['tagWND']]],
                    'spwndChild': [0x70, ['pointer64', ['tagWND']]],
                    'lpfnWndProc': [0xa0, ['pointer64', ['void']]],
                    'pcls': [0xa8, ['pointer64', ['tagCLS']]],
                    'strName': [0xe8, ['_LARGE_UNICODE_STRING']],
                }],
            })


# Volatility
# Copyright (C) 2007-2013 Volatility Foundation
#
# This file is part of Volatility.
#
# Volatility is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# Volatility is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Volatility.  If not, see <http://www.gnu.org/licenses/>.
#

"""
@author:       Andrew Case
@license:      GNU General Public License 2.0
@contact:      atcuno@gmail.com
@organization: 
"""
import volatility.obj   as obj
import volatility.debug as debug
import volatility.plugins.mac.common as common

class mac_vfsevents(common.AbstractMacCommand):
    """ Lists processes filtering file system events """

    def calculate(self):
        common.set_plugin_members(self)

        if not self.addr_space.profile.obj_has_member("fs_event_watcher", "proc_name"):
            debug.error("This plugin only supports OS X >= 10.8.2. Please file a bug if you are running against a version matching this criteria.")

        event_types = ["CREATE_FILE", "DELETE", "STAT_CHANGED", "RENAME", "CONTENT_MODIFIED", "EXCHANGE", "FINDER_INFO_CHANGED", "CREATE_DIR", "CHOWN"]
        event_types = event_types + ["XATTR_MODIFIED", "XATTR_REMOVED", "DOCID_CREATED", "DOCID_CHANGED"]

        table_addr = self.addr_space.profile.get_symbol("_watcher_table")
    
        arr = obj.Object(theType = "Array", targetType = "Pointer", count = 8, vm = self.addr_space, offset = table_addr)

        for watcher_addr in arr:
            if not watcher_addr.is_valid():
                continue

            watcher = watcher_addr.dereference_as("fs_event_watcher")

            name = self.addr_space.read(watcher.proc_name.obj_offset, 33)
            if name:
                idx = name.find("\x00")
                if idx != -1:
                    name = name[:idx]

            events = ""
            event_arr = obj.Object(theType = "Array", targetType = "unsigned char", offset = watcher.event_list.v(), count = 13, vm = self.addr_space)
            for (i, event) in enumerate(event_arr):
                if event == 1:
                    events = events + event_types[i] + ", "  

            if len(events) and events[-1] == " " and events[-2] == ",":
                events = events[:-2]

            yield watcher_addr, name, watcher.pid, events

    def render_text(self, outfd, data):
        self.table_header(outfd, [("Offset", "[addrpad]"),
                          ("Name", "20"),
                          ("Pid", "8"),
                          ("Events", "")])
        
        for (addr, name, pid, events) in data:
            self.table_row(outfd, addr, name, pid, events)



syscalls = [
    [
        "stub_UserGetOwnerTransformedMonitorRect",
        "stub_UserYieldTask",
        "stub_UserSetSensorPresence",
        "NtUserGetThreadState",
        "NtUserPeekMessage",
        "NtUserCallOneParam",
        "NtUserGetKeyState",
        "NtUserInvalidateRect",
        "NtUserCallNoParam",
        "NtUserGetMessage",
        "NtUserMessageCall",
        "NtGdiBitBlt",
        "NtGdiGetCharSet",
        "NtUserGetDC",
        "NtGdiSelectBitmap",
        "NtUserWaitMessage",
        "NtUserTranslateMessage",
        "NtUserGetProp",
        "NtUserPostMessage",
        "NtUserQueryWindow",
        "stub_UserTranslateAccelerator",
        "NtGdiFlush",
        "NtUserRedrawWindow",
        "NtUserWindowFromPoint",
        "NtUserCallMsgFilter",
        "NtUserValidateTimerCallback",
        "NtUserBeginPaint",
        "NtUserSetTimer",
        "NtUserEndPaint",
        "NtUserSetCursor",
        "NtUserKillTimer",
        "NtUserBuildHwndList",
        "NtUserSelectPalette",
        "NtUserCallNextHookEx",
        "NtUserHideCaret",
        "NtGdiIntersectClipRect",
        "NtUserCallHwndLock",
        "NtUserGetProcessWindowStation",
        "NtGdiDeleteObjectApp",
        "NtUserSetWindowPos",
        "NtUserShowCaret",
        "stub_UserEndDeferWindowPosEx",
        "NtUserCallHwndParamLock",
        "NtUserVkKeyScanEx",
        "NtGdiSetDIBitsToDeviceInternal",
        "NtUserCallTwoParam",
        "NtGdiGetRandomRgn",
        "stub_UserCopyAcceleratorTable",
        "NtUserNotifyWinEvent",
        "NtGdiExtSelectClipRgn",
        "NtUserIsClipboardFormatAvailable",
        "NtUserSetScrollInfo",
        "NtGdiStretchBlt",
        "NtUserCreateCaret",
        "NtGdiRectVisible",
        "NtGdiCombineRgn",
        "NtGdiGetDCObject",
        "NtUserDispatchMessage",
        "NtUserRegisterWindowMessage",
        "NtGdiExtTextOutW",
        "NtGdiSelectFont",
        "NtGdiRestoreDC",
        "NtGdiSaveDC",
        "NtUserGetForegroundWindow",
        "stub_UserShowScrollBar",
        "NtUserFindExistingCursorIcon",
        "NtGdiGetDCDword",
        "NtGdiGetRegionData",
        "NtGdiLineTo",
        "NtUserSystemParametersInfo",
        "NtGdiGetAppClipBox",
        "NtUserGetAsyncKeyState",
        "NtUserGetCPD",
        "NtUserRemoveProp",
        "NtGdiDoPalette",
        "NtGdiPolyPolyDraw",
        "NtUserSetCapture",
        "NtUserEnumDisplayMonitors",
        "NtGdiCreateCompatibleBitmap",
        "NtUserSetProp",
        "NtGdiGetTextCharsetInfo",
        "stub_UserSBGetParms",
        "NtUserGetIconInfo",
        "stub_UserExcludeUpdateRgn",
        "NtUserSetFocus",
        "NtGdiExtGetObjectW",
        "NtUserGetUpdateRect",
        "NtGdiCreateCompatibleDC",
        "NtUserGetClipboardSequenceNumber",
        "NtGdiCreatePen",
        "NtUserShowWindow",
        "NtUserGetKeyboardLayoutList",
        "NtGdiPatBlt",
        "NtUserMapVirtualKeyEx",
        "NtUserSetWindowLong",
        "NtGdiHfontCreate",
        "NtUserMoveWindow",
        "NtUserPostThreadMessage",
        "NtUserDrawIconEx",
        "NtUserGetSystemMenu",
        "NtGdiDrawStream",
        "NtUserInternalGetWindowText",
        "NtUserGetWindowDC",
        "stub_GdiD3dDrawPrimitives2",
        "stub_GdiInvertRgn",
        "NtGdiGetRgnBox",
        "NtGdiGetAndSetDCDword",
        "stub_GdiMaskBlt",
        "NtGdiGetWidthTable",
        "NtUserScrollDC",
        "NtUserGetObjectInformation",
        "NtGdiCreateBitmap",
        "NtUserFindWindowEx",
        "NtGdiPolyPatBlt",
        "NtUserUnhookWindowsHookEx",
        "stub_GdiGetNearestColor",
        "NtGdiTransformPoints",
        "NtGdiGetDCPoint",
        "stub_GdiCreateDIBBrush",
        "NtGdiGetTextMetricsW",
        "NtUserCreateWindowEx",
        "NtUserSetParent",
        "NtUserGetKeyboardState",
        "NtUserToUnicodeEx",
        "NtUserGetControlBrush",
        "NtUserGetClassName",
        "NtGdiAlphaBlend",
        "stub_GdiDdBlt",
        "NtGdiOffsetRgn",
        "NtUserDefSetText",
        "NtGdiGetTextFaceW",
        "NtGdiStretchDIBitsInternal",
        "NtUserSendInput",
        "NtUserGetThreadDesktop",
        "NtGdiCreateRectRgn",
        "NtGdiGetDIBitsInternal",
        "stub_UserGetUpdateRgn",
        "NtGdiDeleteClientObj",
        "NtUserGetIconSize",
        "NtUserFillWindow",
        "NtGdiExtCreateRegion",
        "NtGdiComputeXformCoefficients",
        "NtUserSetWindowsHookEx",
        "stub_UserNotifyProcessCreate",
        "stub_GdiUnrealizeObject",
        "NtUserGetTitleBarInfo",
        "NtGdiRectangle",
        "NtUserSetThreadDesktop",
        "NtUserGetDCEx",
        "stub_UserGetScrollBarInfo",
        "stub_GdiGetTextExtent",
        "NtUserSetWindowFNID",
        "NtGdiSetLayout",
        "NtUserCalcMenuBar",
        "NtUserThunkedMenuItemInfo",
        "NtGdiExcludeClipRect",
        "NtGdiCreateDIBSection",
        "NtGdiGetDCforBitmap",
        "NtUserDestroyCursor",
        "NtUserDestroyWindow",
        "NtUserCallHwndParam",
        "NtGdiCreateDIBitmapInternal",
        "stub_UserOpenWindowStation",
        "stub_GdiDdDeleteSurfaceObject",
        "stub_GdiDdCanCreateSurface",
        "stub_GdiDdCreateSurface",
        "NtUserSetCursorIconData",
        "stub_GdiDdDestroySurface",
        "NtUserCloseDesktop",
        "NtUserOpenDesktop",
        "stub_UserSetProcessWindowStation",
        "NtUserGetAtomName",
        "stub_GdiDdResetVisrgn",
        "NtGdiExtCreatePen",
        "NtGdiCreatePaletteInternal",
        "stub_GdiSetBrushOrg",
        "NtUserBuildNameList",
        "NtGdiSetPixel",
        "NtUserRegisterClassExWOW",
        "NtGdiCreatePatternBrushInternal",
        "NtUserGetAncestor",
        "NtGdiGetOutlineTextMetricsInternalW",
        "NtGdiSetBitmapBits",
        "NtUserCloseWindowStation",
        "NtUserGetDoubleClickTime",
        "stub_UserEnableScrollBar",
        "NtGdiCreateSolidBrush",
        "NtUserGetClassInfoEx",
        "NtGdiCreateClientObj",
        "NtUserUnregisterClass",
        "NtUserDeleteMenu",
        "NtGdiRectInRegion",
        "stub_UserScrollWindowEx",
        "NtGdiGetPixel",
        "NtUserSetClassLong",
        "NtUserGetMenuBarInfo",
        "stub_GdiDdCreateSurfaceEx",
        "stub_GdiDdCreateSurfaceObject",
        "stub_GdiGetNearestPaletteIndex",
        "stub_GdiDdLockD3D",
        "stub_GdiDdUnlockD3D",
        "NtGdiGetCharWidthW",
        "stub_UserInvalidateRgn",
        "stub_UserGetClipboardOwner",
        "NtUserSetWindowRgn",
        "NtUserBitBltSysBmp",
        "stub_GdiGetCharWidthInfo",
        "NtUserValidateRect",
        "NtUserCloseClipboard",
        "NtUserOpenClipboard",
        "stub_UserSetClipboardData",
        "NtUserEnableMenuItem",
        "NtUserAlterWindowStyle",
        "NtGdiFillRgn",
        "NtUserGetWindowPlacement",
        "NtGdiModifyWorldTransform",
        "NtGdiGetFontData",
        "NtUserGetOpenClipboardWindow",
        "NtUserSetThreadState",
        "NtGdiOpenDCW",
        "NtUserTrackMouseEvent",
        "NtGdiGetTransform",
        "NtUserDestroyMenu",
        "NtGdiGetBitmapBits",
        "stub_UserConsoleControl",
        "NtUserSetActiveWindow",
        "stub_UserSetInformationThread",
        "stub_UserSetWindowPlacement",
        "NtUserGetControlColor",
        "NtGdiSetMetaRgn",
        "stub_GdiSetMiterLimit",
        "stub_GdiSetVirtualResolution",
        "stub_GdiGetRasterizerCaps",
        "stub_UserSetWindowWord",
        "NtUserGetClipboardFormatName",
        "NtUserRealInternalGetMessage",
        "stub_UserCreateLocalMemHandle",
        "NtUserAttachThreadInput",
        "NtGdiCreateHalftonePalette",
        "stub_UserPaintMenuBar",
        "NtUserSetKeyboardState",
        "stub_GdiCombineTransform",
        "NtUserCreateAcceleratorTable",
        "NtUserGetCursorFrameInfo",
        "stub_UserGetAltTabInfo",
        "NtUserGetCaretBlinkTime",
        "NtGdiQueryFontAssocInfo",
        "stub_UserProcessConnect",
        "NtUserEnumDisplayDevices",
        "stub_UserEmptyClipboard",
        "NtUserGetClipboardData",
        "NtUserRemoveMenu",
        "NtGdiSetBoundsRect",
        "stub_GdiGetBitmapDimension",
        "stub_UserConvertMemHandle",
        "NtUserDestroyAcceleratorTable",
        "NtUserGetGUIThreadInfo",
        "stub_GdiCloseFigure",
        "stub_UserSetWindowsHookAW",
        "stub_UserSetMenuDefaultItem",
        "NtUserCheckMenuItem",
        "NtUserSetWinEventHook",
        "NtUserUnhookWinEvent",
        "stub_UserLockWindowUpdate",
        "stub_UserSetSystemMenu",
        "NtUserThunkedMenuInfo",
        "NtGdiBeginPath",
        "NtGdiEndPath",
        "NtGdiFillPath",
        "NtUserCallHwnd",
        "stub_UserDdeInitialize",
        "NtUserModifyUserStartupInfoFlags",
        "NtUserCountClipboardFormats",
        "stub_GdiAddFontMemResourceEx",
        "stub_GdiEqualRgn",
        "stub_GdiGetSystemPaletteUse",
        "stub_GdiRemoveFontMemResourceEx",
        "NtUserEnumDisplaySettings",
        "stub_UserPaintDesktop",
        "stub_GdiExtEscape",
        "stub_GdiSetBitmapDimension",
        "stub_GdiSetFontEnumeration",
        "NtUserChangeClipboardChain",
        "NtUserSetClipboardViewer",
        "stub_UserShowWindowAsync",
        "stub_GdiCreateColorSpace",
        "stub_GdiDeleteColorSpace",
        "NtUserActivateKeyboardLayout",
        "NtBindCompositionSurface",
        "stub_CompositionInputThread",
        "NtCompositionSetDropTarget",
        "NtCreateCompositionInputSink",
        "NtCreateCompositionSurfaceHandle",
        "stub_CreateImplicitCompositionInputSink",
        "NtDCompositionAddCrossDeviceVisualChild",
        "stub_DCompositionBeginFrame",
        "NtDCompositionCommitChannel",
        "stub_DCompositionConfirmFrame",
        "stub_DCompositionConnectPipe",
        "NtDCompositionCreateAndBindSharedSection",
        "NtDCompositionCreateChannel",
        "stub_DCompositionCreateConnection",
        "stub_DCompositionCreateDwmChannel",
        "NtDCompositionCurrentBatchId",
        "NtDCompositionDestroyChannel",
        "stub_DCompositionDestroyConnection",
        "stub_DCompositionDiscardFrame",
        "stub_DCompositionDuplicateHandleToProcess",
        "stub_DCompositionDuplicateSwapchainHandleToDwm",
        "stub_DCompositionEnableDDASupport",
        "stub_DCompositionEnableMMCSS",
        "stub_DCompositionGetChannels",
        "stub_DCompositionGetConnectionBatch",
        "NtDCompositionGetDeletedResources",
        "stub_DCompositionGetFrameLegacyTokens",
        "NtDCompositionGetFrameStatistics",
        "stub_DCompositionGetFrameSurfaceUpdates",
        "NtDCompositionProcessChannelBatchBuffer",
        "stub_DCompositionReferenceSharedResourceOnDwmChannel",
        "stub_DCompositionRegisterThumbnailVisual",
        "stub_DCompositionRegisterVirtualDesktopVisual",
        "NtDCompositionReleaseAllResources",
        "NtDCompositionRemoveCrossDeviceVisualChild",
        "stub_DCompositionRetireFrame",
        "NtDCompositionSetChannelCallbackId",
        "NtDCompositionSetChannelCommitCompletionEvent",
        "stub_DCompositionSetDebugCounter",
        "stub_DCompositionSubmitDWMBatch",
        "NtDCompositionSynchronize",
        "NtDCompositionTelemetryAnimationScenarioBegin",
        "NtDCompositionTelemetryAnimationScenarioReference",
        "NtDCompositionTelemetryAnimationScenarioUnreference",
        "NtDCompositionTelemetrySetApplicationId",
        "NtDCompositionTelemetryTouchInteractionBegin",
        "NtDCompositionTelemetryTouchInteractionEnd",
        "NtDCompositionTelemetryTouchInteractionUpdate",
        "stub_DCompositionUpdatePointerCapture",
        "NtDCompositionWaitForChannel",
        "stub_DesktopCaptureBits",
        "stub_DuplicateCompositionInputSink",
        "stub_GdiAbortDoc",
        "stub_GdiAbortPath",
        "stub_GdiAddEmbFontToDC",
        "stub_GdiAddFontResourceW",
        "stub_GdiAddRemoteFontToDC",
        "stub_GdiAddRemoteMMInstanceToDC",
        "stub_GdiAngleArc",
        "NtGdiAnyLinkedFonts",
        "stub_GdiArcInternal",
        "stub_GdiBRUSHOBJ_DeleteRbrush",
        "stub_GdiBRUSHOBJ_hGetColorTransform",
        "stub_GdiBRUSHOBJ_pvAllocRbrush",
        "stub_GdiBRUSHOBJ_pvGetRbrush",
        "stub_GdiBRUSHOBJ_ulGetBrushColor",
        "NtGdiBeginGdiRendering",
        "stub_GdiCLIPOBJ_bEnum",
        "stub_GdiCLIPOBJ_cEnumStart",
        "stub_GdiCLIPOBJ_ppoGetPath",
        "stub_GdiCancelDC",
        "stub_GdiChangeGhostFont",
        "stub_GdiCheckBitmapBits",
        "stub_GdiClearBitmapAttributes",
        "stub_GdiClearBrushAttributes",
        "stub_GdiColorCorrectPalette",
        "NtGdiConfigureOPMProtectedOutput",
        "stub_GdiConvertMetafileRect",
        "stub_GdiCreateBitmapFromDxSurface",
        "NtGdiCreateBitmapFromDxSurface2",
        "stub_GdiCreateColorTransform",
        "stub_GdiCreateEllipticRgn",
        "stub_GdiCreateHatchBrushInternal",
        "stub_GdiCreateMetafileDC",
        "NtGdiCreateOPMProtectedOutput",
        "NtGdiCreateOPMProtectedOutputs",
        "NtGdiCreateRoundRectRgn",
        "stub_GdiCreateServerMetaFile",
        "stub_GdiCreateSessionMappedDIBSection",
        "stub_GdiD3dContextCreate",
        "stub_GdiD3dContextDestroy",
        "stub_GdiD3dContextDestroyAll",
        "stub_GdiD3dValidateTextureStageState",
        "stub_GdiDDCCIGetCapabilitiesString",
        "stub_GdiDDCCIGetCapabilitiesStringLength",
        "stub_GdiDDCCIGetTimingReport",
        "stub_GdiDDCCIGetVCPFeature",
        "stub_GdiDDCCISaveCurrentSettings",
        "stub_GdiDDCCISetVCPFeature",
        "stub_GdiDdAddAttachedSurface",
        "stub_GdiDdAlphaBlt",
        "stub_GdiDdAttachSurface",
        "stub_GdiDdBeginMoCompFrame",
        "stub_GdiDdCanCreateD3DBuffer",
        "stub_GdiDdColorControl",
        "stub_GdiDdCreateD3DBuffer",
        "stub_GdiDdCreateDirectDrawObject",
        "stub_GdiDdCreateFullscreenSprite",
        "stub_GdiDdCreateMoComp",
        "stub_GdiDdDDIAbandonSwapChain",
        "stub_GdiDdDDIAcquireKeyedMutex",
        "NtGdiDdDDIAcquireKeyedMutex2",
        "stub_GdiDdDDIAcquireSwapChain",
        "stub_GdiDdDDIAdjustFullscreenGamma",
        "NtGdiDdDDICacheHybridQueryValue",
        "stub_GdiDdDDIChangeVideoMemoryReservation",
        "stub_GdiDdDDICheckExclusiveOwnership",
        "stub_GdiDdDDICheckMonitorPowerState",
        "stub_GdiDdDDICheckMultiPlaneOverlaySupport",
        "stub_GdiDdDDICheckMultiPlaneOverlaySupport2",
        "stub_GdiDdDDICheckMultiPlaneOverlaySupport3",
        "stub_GdiDdDDICheckOcclusion",
        "stub_GdiDdDDICheckSharedResourceAccess",
        "stub_GdiDdDDICheckVidPnExclusiveOwnership",
        "NtGdiDdDDICloseAdapter",
        "NtGdiDdDDIConfigureSharedResource",
        "NtGdiDdDDICreateAllocation",
        "NtGdiDdDDICreateContext",
        "NtGdiDdDDICreateContextVirtual",
        "NtGdiDdDDICreateDCFromMemory",
        "NtGdiDdDDICreateDevice",
        "stub_GdiDdDDICreateKeyedMutex",
        "NtGdiDdDDICreateKeyedMutex2",
        "stub_GdiDdDDICreateOutputDupl",
        "stub_GdiDdDDICreateOverlay",
        "NtGdiDdDDICreatePagingQueue",
        "stub_GdiDdDDICreateSwapChain",
        "NtGdiDdDDICreateSynchronizationObject",
        "NtGdiDdDDIDestroyAllocation",
        "NtGdiDdDDIDestroyAllocation2",
        "NtGdiDdDDIDestroyContext",
        "NtGdiDdDDIDestroyDCFromMemory",
        "NtGdiDdDDIDestroyDevice",
        "NtGdiDdDDIDestroyKeyedMutex",
        "stub_GdiDdDDIDestroyOutputDupl",
        "stub_GdiDdDDIDestroyOverlay",
        "NtGdiDdDDIDestroyPagingQueue",
        "NtGdiDdDDIDestroySynchronizationObject",
        "NtGdiDdDDIEnumAdapters",
        "NtGdiDdDDIEnumAdapters2",
        "NtGdiDdDDIEscape",
        "NtGdiDdDDIEvict",
        "stub_GdiDdDDIFlipOverlay",
        "NtGdiDdDDIFlushHeapTransitions",
        "NtGdiDdDDIFreeGpuVirtualAddress",
        "NtGdiDdDDIGetCachedHybridQueryValue",
        "stub_GdiDdDDIGetContextInProcessSchedulingPriority",
        "NtGdiDdDDIGetContextSchedulingPriority",
        "stub_GdiDdDDIGetDWMVerticalBlankEvent",
        "NtGdiDdDDIGetDeviceState",
        "NtGdiDdDDIGetDisplayModeList",
        "stub_GdiDdDDIGetMultisampleMethodList",
        "stub_GdiDdDDIGetOverlayState",
        "stub_GdiDdDDIGetPresentHistory",
        "stub_GdiDdDDIGetPresentQueueEvent",
        "stub_GdiDdDDIGetProcessSchedulingPriorityClass",
        "stub_GdiDdDDIGetResourcePresentPrivateDriverData",
        "stub_GdiDdDDIGetRuntimeData",
        "stub_GdiDdDDIGetScanLine",
        "stub_GdiDdDDIGetSetSwapChainMetadata",
        "NtGdiDdDDIGetSharedPrimaryHandle",
        "stub_GdiDdDDIGetSharedResourceAdapterLuid",
        "stub_GdiDdDDIInvalidateActiveVidPn",
        "stub_GdiDdDDIInvalidateCache",
        "NtGdiDdDDILock",
        "NtGdiDdDDILock2",
        "NtGdiDdDDIMakeResident",
        "NtGdiDdDDIMapGpuVirtualAddress",
        "NtGdiDdDDIMarkDeviceAsError",
        "stub_GdiDdDDINetDispGetNextChunkInfo",
        "stub_GdiDdDDINetDispQueryMiracastDisplayDeviceStatus",
        "NtGdiDdDDINetDispQueryMiracastDisplayDeviceSupport",
        "stub_GdiDdDDINetDispStartMiracastDisplayDevice",
        "stub_GdiDdDDINetDispStopMiracastDisplayDevice",
        "stub_GdiDdDDINetDispStopSessions",
        "NtGdiDdDDIOfferAllocations",
        "NtGdiDdDDIOpenAdapterFromDeviceName",
        "NtGdiDdDDIOpenAdapterFromHdc",
        "NtGdiDdDDIOpenAdapterFromLuid",
        "stub_GdiDdDDIOpenKeyedMutex",
        "NtGdiDdDDIOpenKeyedMutex2",
        "stub_GdiDdDDIOpenNtHandleFromName",
        "NtGdiDdDDIOpenResource",
        "NtGdiDdDDIOpenResourceFromNtHandle",
        "stub_GdiDdDDIOpenSwapChain",
        "NtGdiDdDDIOpenSyncObjectFromNtHandle",
        "stub_GdiDdDDIOpenSyncObjectFromNtHandle2",
        "stub_GdiDdDDIOpenSyncObjectNtHandleFromName",
        "NtGdiDdDDIOpenSynchronizationObject",
        "stub_GdiDdDDIOutputDuplGetFrameInfo",
        "stub_GdiDdDDIOutputDuplGetMetaData",
        "stub_GdiDdDDIOutputDuplGetPointerShapeData",
        "stub_GdiDdDDIOutputDuplPresent",
        "stub_GdiDdDDIOutputDuplReleaseFrame",
        "stub_GdiDdDDIPinDirectFlipResources",
        "NtGdiDdDDIPollDisplayChildren",
        "NtGdiDdDDIPresent",
        "stub_GdiDdDDIPresentMultiPlaneOverlay",
        "stub_GdiDdDDIPresentMultiPlaneOverlay2",
        "stub_GdiDdDDIPresentMultiPlaneOverlay3",
        "NtGdiDdDDIQueryAdapterInfo",
        "stub_GdiDdDDIQueryAllocationResidency",
        "stub_GdiDdDDIQueryClockCalibration",
        "stub_GdiDdDDIQueryFSEBlock",
        "stub_GdiDdDDIQueryProcessOfferInfo",
        "stub_GdiDdDDIQueryRemoteVidPnSourceFromGdiDisplayName",
        "NtGdiDdDDIQueryResourceInfo",
        "NtGdiDdDDIQueryResourceInfoFromNtHandle",
        "stub_GdiDdDDIQueryStatistics",
        "stub_GdiDdDDIQueryVidPnExclusiveOwnership",
        "NtGdiDdDDIQueryVideoMemoryInfo",
        "NtGdiDdDDIReclaimAllocations",
        "NtGdiDdDDIReclaimAllocations2",
        "stub_GdiDdDDIReleaseKeyedMutex",
        "NtGdiDdDDIReleaseKeyedMutex2",
        "stub_GdiDdDDIReleaseProcessVidPnSourceOwners",
        "stub_GdiDdDDIReleaseSwapChain",
        "NtGdiDdDDIRender",
        "NtGdiDdDDIReserveGpuVirtualAddress",
        "NtGdiDdDDISetAllocationPriority",
        "stub_GdiDdDDISetContextInProcessSchedulingPriority",
        "NtGdiDdDDISetContextSchedulingPriority",
        "stub_GdiDdDDISetDisplayMode",
        "stub_GdiDdDDISetDisplayPrivateDriverFormat",
        "stub_GdiDdDDISetDodIndirectSwapchain",
        "stub_GdiDdDDISetFSEBlock",
        "stub_GdiDdDDISetGammaRamp",
        "NtGdiDdDDISetHwProtectionTeardownRecovery",
        "stub_GdiDdDDISetProcessSchedulingPriorityClass",
        "NtGdiDdDDISetQueuedLimit",
        "stub_GdiDdDDISetStablePowerState",
        "stub_GdiDdDDISetStereoEnabled",
        "stub_GdiDdDDISetSyncRefreshCountWaitTarget",
        "stub_GdiDdDDISetVidPnSourceHwProtection",
        "stub_GdiDdDDISetVidPnSourceOwner",
        "stub_GdiDdDDISetVidPnSourceOwner1",
        "NtGdiDdDDIShareObjects",
        "stub_GdiDdDDISharedPrimaryLockNotification",
        "stub_GdiDdDDISharedPrimaryUnLockNotification",
        "NtGdiDdDDISignalSynchronizationObject",
        "stub_GdiDdDDISignalSynchronizationObjectFromCpu",
        "NtGdiDdDDISignalSynchronizationObjectFromGpu",
        "NtGdiDdDDISignalSynchronizationObjectFromGpu2",
        "NtGdiDdDDISubmitCommand",
        "stub_GdiDdDDITrimProcessCommitment",
        "NtGdiDdDDIUnlock",
        "NtGdiDdDDIUnlock2",
        "stub_GdiDdDDIUnpinDirectFlipResources",
        "stub_GdiDdDDIUpdateAllocationProperty",
        "stub_GdiDdDDIUpdateGpuVirtualAddress",
        "stub_GdiDdDDIUpdateOverlay",
        "stub_GdiDdDDIWaitForIdle",
        "NtGdiDdDDIWaitForSynchronizationObject",
        "NtGdiDdDDIWaitForSynchronizationObjectFromCpu",
        "NtGdiDdDDIWaitForSynchronizationObjectFromGpu",
        "NtGdiDdDDIWaitForVerticalBlankEvent",
        "stub_GdiDdDDIWaitForVerticalBlankEvent2",
        "stub_GdiDdDeleteDirectDrawObject",
        "stub_GdiDdDestroyD3DBuffer",
        "stub_GdiDdDestroyFullscreenSprite",
        "stub_GdiDdDestroyMoComp",
        "stub_GdiDdEndMoCompFrame",
        "stub_GdiDdFlip",
        "stub_GdiDdFlipToGDISurface",
        "stub_GdiDdGetAvailDriverMemory",
        "stub_GdiDdGetBltStatus",
        "stub_GdiDdGetDC",
        "stub_GdiDdGetDriverInfo",
        "stub_GdiDdGetDriverState",
        "stub_GdiDdGetDxHandle",
        "stub_GdiDdGetFlipStatus",
        "stub_GdiDdGetInternalMoCompInfo",
        "stub_GdiDdGetMoCompBuffInfo",
        "stub_GdiDdGetMoCompFormats",
        "stub_GdiDdGetMoCompGuids",
        "stub_GdiDdGetScanLine",
        "stub_GdiDdLock",
        "stub_GdiDdNotifyFullscreenSpriteUpdate",
        "stub_GdiDdQueryDirectDrawObject",
        "stub_GdiDdQueryMoCompStatus",
        "stub_GdiDdQueryVisRgnUniqueness",
        "stub_GdiDdReenableDirectDrawObject",
        "stub_GdiDdReleaseDC",
        "stub_GdiDdRenderMoComp",
        "stub_GdiDdSetColorKey",
        "stub_GdiDdSetExclusiveMode",
        "stub_GdiDdSetGammaRamp",
        "stub_GdiDdSetOverlayPosition",
        "stub_GdiDdUnattachSurface",
        "stub_GdiDdUnlock",
        "stub_GdiDdUpdateOverlay",
        "stub_GdiDdWaitForVerticalBlank",
        "stub_GdiDeleteColorTransform",
        "stub_GdiDescribePixelFormat",
        "NtGdiDestroyOPMProtectedOutput",
        "stub_GdiDestroyPhysicalMonitor",
        "stub_GdiDoBanding",
        "stub_GdiDrawEscape",
        "stub_GdiDvpAcquireNotification",
        "stub_GdiDvpCanCreateVideoPort",
        "stub_GdiDvpColorControl",
        "stub_GdiDvpCreateVideoPort",
        "stub_GdiDvpDestroyVideoPort",
        "stub_GdiDvpFlipVideoPort",
        "stub_GdiDvpGetVideoPortBandwidth",
        "stub_GdiDvpGetVideoPortConnectInfo",
        "stub_GdiDvpGetVideoPortField",
        "stub_GdiDvpGetVideoPortFlipStatus",
        "stub_GdiDvpGetVideoPortInputFormats",
        "stub_GdiDvpGetVideoPortLine",
        "stub_GdiDvpGetVideoPortOutputFormats",
        "stub_GdiDvpGetVideoSignalStatus",
        "stub_GdiDvpReleaseNotification",
        "stub_GdiDvpUpdateVideoPort",
        "stub_GdiDvpWaitForVideoPortSync",
        "stub_GdiDwmCreatedBitmapRemotingOutput",
        "stub_GdiDxgGenericThunk",
        "NtGdiEllipse",
        "stub_GdiEnableEudc",
        "stub_GdiEndDoc",
        "NtGdiEndGdiRendering",
        "stub_GdiEndPage",
        "stub_GdiEngAlphaBlend",
        "stub_GdiEngAssociateSurface",
        "stub_GdiEngBitBlt",
        "stub_GdiEngCheckAbort",
        "stub_GdiEngComputeGlyphSet",
        "stub_GdiEngCopyBits",
        "stub_GdiEngCreateBitmap",
        "stub_GdiEngCreateClip",
        "stub_GdiEngCreateDeviceBitmap",
        "stub_GdiEngCreateDeviceSurface",
        "stub_GdiEngCreatePalette",
        "stub_GdiEngDeleteClip",
        "stub_GdiEngDeletePalette",
        "stub_GdiEngDeletePath",
        "stub_GdiEngDeleteSurface",
        "stub_GdiEngEraseSurface",
        "stub_GdiEngFillPath",
        "stub_GdiEngGradientFill",
        "stub_GdiEngLineTo",
        "stub_GdiEngLockSurface",
        "stub_GdiEngMarkBandingSurface",
        "stub_GdiEngPaint",
        "stub_GdiEngPlgBlt",
        "stub_GdiEngStretchBlt",
        "stub_GdiEngStretchBltROP",
        "stub_GdiEngStrokeAndFillPath",
        "stub_GdiEngStrokePath",
        "stub_GdiEngTextOut",
        "stub_GdiEngTransparentBlt",
        "stub_GdiEngUnlockSurface",
        "NtGdiEnumFonts",
        "stub_GdiEnumObjects",
        "stub_GdiEudcLoadUnloadLink",
        "stub_GdiExtFloodFill",
        "stub_GdiFONTOBJ_cGetAllGlyphHandles",
        "stub_GdiFONTOBJ_cGetGlyphs",
        "stub_GdiFONTOBJ_pQueryGlyphAttrs",
        "stub_GdiFONTOBJ_pfdg",
        "stub_GdiFONTOBJ_pifi",
        "stub_GdiFONTOBJ_pvTrueTypeFontFile",
        "stub_GdiFONTOBJ_pxoGetXform",
        "stub_GdiFONTOBJ_vGetInfo",
        "stub_GdiFlattenPath",
        "NtGdiFontIsLinked",
        "stub_GdiForceUFIMapping",
        "NtGdiFrameRgn",
        "stub_GdiFullscreenControl",
        "NtGdiGetBoundsRect",
        "stub_GdiGetCOPPCompatibleOPMInformation",
        "stub_GdiGetCertificate",
        "NtGdiGetCertificateByHandle",
        "stub_GdiGetCertificateSize",
        "NtGdiGetCertificateSizeByHandle",
        "NtGdiGetCharABCWidthsW",
        "stub_GdiGetCharacterPlacementW",
        "stub_GdiGetColorAdjustment",
        "stub_GdiGetColorSpaceforBitmap",
        "NtGdiGetCurrentDpiInfo",
        "NtGdiGetDeviceCaps",
        "stub_GdiGetDeviceCapsAll",
        "stub_GdiGetDeviceGammaRamp",
        "NtGdiGetDeviceWidth",
        "stub_GdiGetDhpdev",
        "stub_GdiGetETM",
        "stub_GdiGetEmbUFI",
        "stub_GdiGetEmbedFonts",
        "NtGdiGetEntry",
        "stub_GdiGetEudcTimeStampEx",
        "stub_GdiGetFontFileData",
        "stub_GdiGetFontFileInfo",
        "stub_GdiGetFontResourceInfoInternalW",
        "stub_GdiGetFontUnicodeRanges",
        "NtGdiGetGlyphIndicesW",
        "stub_GdiGetGlyphIndicesWInternal",
        "stub_GdiGetGlyphOutline",
        "stub_GdiGetKerningPairs",
        "stub_GdiGetLinkedUFIs",
        "stub_GdiGetMiterLimit",
        "NtGdiGetMonitorID",
        "NtGdiGetNumberOfPhysicalMonitors",
        "NtGdiGetOPMInformation",
        "NtGdiGetOPMRandomNumber",
        "stub_GdiGetObjectBitmapHandle",
        "stub_GdiGetPath",
        "stub_GdiGetPerBandInfo",
        "stub_GdiGetPhysicalMonitorDescription",
        "stub_GdiGetPhysicalMonitors",
        "stub_GdiGetProcessSessionFonts",
        "stub_GdiGetPublicFontTableChangeCookie",
        "NtGdiGetRealizationInfo",
        "stub_GdiGetServerMetaFileBits",
        "stub_GdiGetSpoolMessage",
        "stub_GdiGetStats",
        "stub_GdiGetStringBitmapW",
        "NtGdiGetSuggestedOPMProtectedOutputArraySize",
        "NtGdiGetTextExtentExW",
        "stub_GdiGetUFI",
        "stub_GdiGetUFIPathname",
        "NtGdiGradientFill",
        "stub_GdiHLSurfGetInformation",
        "stub_GdiHLSurfSetInformation",
        "stub_GdiHT_Get8BPPFormatPalette",
        "stub_GdiHT_Get8BPPMaskPalette",
        "stub_GdiIcmBrushInfo",
        "NtGdiInit",
        "stub_GdiInitSpool",
        "stub_GdiMakeFontDir",
        "stub_GdiMakeInfoDC",
        "stub_GdiMakeObjectUnXferable",
        "stub_GdiMakeObjectXferable",
        "stub_GdiMirrorWindowOrg",
        "stub_GdiMonoBitmap",
        "stub_GdiMoveTo",
        "stub_GdiOffsetClipRgn",
        "stub_GdiPATHOBJ_bEnum",
        "stub_GdiPATHOBJ_bEnumClipLines",
        "stub_GdiPATHOBJ_vEnumStart",
        "stub_GdiPATHOBJ_vEnumStartClipLines",
        "stub_GdiPATHOBJ_vGetBounds",
        "stub_GdiPathToRegion",
        "stub_GdiPlgBlt",
        "stub_GdiPolyDraw",
        "stub_GdiPolyTextOutW",
        "stub_GdiPtInRegion",
        "stub_GdiPtVisible",
        "stub_GdiQueryFonts",
        "stub_GdiRemoveFontResourceW",
        "stub_GdiRemoveMergeFont",
        "stub_GdiResetDC",
        "stub_GdiResizePalette",
        "NtGdiRoundRect",
        "stub_GdiSTROBJ_bEnum",
        "stub_GdiSTROBJ_bEnumPositionsOnly",
        "stub_GdiSTROBJ_bGetAdvanceWidths",
        "stub_GdiSTROBJ_dwGetCodePage",
        "stub_GdiSTROBJ_vEnumStart",
        "stub_GdiScaleViewportExtEx",
        "stub_GdiScaleWindowExtEx",
        "stub_GdiSelectBrush",
        "NtGdiSelectClipPath",
        "stub_GdiSelectPen",
        "stub_GdiSetBitmapAttributes",
        "stub_GdiSetBrushAttributes",
        "stub_GdiSetColorAdjustment",
        "stub_GdiSetColorSpace",
        "stub_GdiSetDeviceGammaRamp",
        "stub_GdiSetFontXform",
        "NtGdiSetIcmMode",
        "stub_GdiSetLinkedUFIs",
        "stub_GdiSetMagicColors",
        "NtGdiSetOPMSigningKeyAndSequenceNumbers",
        "stub_GdiSetPUMPDOBJ",
        "stub_GdiSetPixelFormat",
        "stub_GdiSetRectRgn",
        "stub_GdiSetSizeDevice",
        "stub_GdiSetSystemPaletteUse",
        "stub_GdiSetTextJustification",
        "NtGdiSetUMPDSandboxState",
        "stub_GdiStartDoc",
        "stub_GdiStartPage",
        "stub_GdiStrokeAndFillPath",
        "NtGdiStrokePath",
        "stub_GdiSwapBuffers",
        "NtGdiTransparentBlt",
        "stub_GdiUMPDEngFreeUserMem",
        "stub_GdiUnloadPrinterDriver",
        "stub_GdiUnmapMemFont",
        "stub_GdiUpdateColors",
        "stub_GdiUpdateTransform",
        "stub_GdiWidenPath",
        "stub_GdiXFORMOBJ_bApplyXform",
        "stub_GdiXFORMOBJ_iGetXform",
        "stub_GdiXLATEOBJ_cGetPalette",
        "stub_GdiXLATEOBJ_hGetColorTransform",
        "stub_GdiXLATEOBJ_iXlate",
        "stub_HWCursorUpdatePointer",
        "stub_NotifyPresentToCompositionSurface",
        "stub_OpenCompositionSurfaceDirtyRegion",
        "stub_OpenCompositionSurfaceSectionInfo",
        "stub_OpenCompositionSurfaceSwapChainHandleInfo",
        "stub_QueryCompositionInputIsImplicit",
        "stub_QueryCompositionInputQueueAndTransform",
        "stub_QueryCompositionInputSink",
        "stub_QueryCompositionInputSinkLuid",
        "stub_QueryCompositionInputSinkViewId",
        "stub_QueryCompositionSurfaceBinding",
        "stub_QueryCompositionSurfaceHDRMetaData",
        "stub_QueryCompositionSurfaceRenderingRealization",
        "NtQueryCompositionSurfaceStatistics",
        "stub_RIMAddInputObserver",
        "stub_RIMGetDevicePreparsedDataLockfree",
        "stub_RIMObserveNextInput",
        "stub_RIMRemoveInputObserver",
        "stub_RIMUpdateInputObserverRegistration",
        "stub_SetCompositionSurfaceAnalogExclusive",
        "stub_SetCompositionSurfaceBufferCompositionModeAndOrientation",
        "stub_SetCompositionSurfaceDirectFlipState",
        "stub_SetCompositionSurfaceHDRMetaData",
        "stub_SetCompositionSurfaceIndependentFlipInfo",
        "stub_SetCompositionSurfaceStatistics",
        "stub_TokenManagerConfirmOutstandingAnalogToken",
        "NtTokenManagerCreateCompositionTokenHandle",
        "stub_TokenManagerGetAnalogExclusiveSurfaceUpdates",
        "stub_TokenManagerGetAnalogExclusiveTokenEvent",
        "stub_TokenManagerOpenSectionAndEvents",
        "stub_TokenManagerThread",
        "NtUnBindCompositionSurface",
        "stub_UpdateInputSinkTransforms",
        "stub_UserAcquireIAMKey",
        "stub_UserAcquireInteractiveControlBackgroundAccess",
        "stub_UserAddClipboardFormatListener",
        "NtUserAssociateInputContext",
        "NtUserAutoPromoteMouseInPointer",
        "stub_UserAutoRotateScreen",
        "stub_UserBlockInput",
        "stub_UserBroadcastThemeChangeEvent",
        "NtUserBuildHimcList",
        "stub_UserBuildPropList",
        "stub_UserCalculatePopupWindowPosition",
        "stub_UserCallHwndOpt",
        "stub_UserCanBrokerForceForeground",
        "stub_UserChangeDisplaySettings",
        "NtUserChangeWindowMessageFilterEx",
        "stub_UserCheckAccessForIntegrityLevel",
        "stub_UserCheckProcessForClipboardAccess",
        "stub_UserCheckProcessSession",
        "stub_UserCheckWindowThreadDesktop",
        "NtUserChildWindowFromPointEx",
        "stub_UserClearForeground",
        "NtUserClipCursor",
        "stub_UserCompositionInputSinkLuidFromPoint",
        "NtUserCreateDCompositionHwndTarget",
        "stub_UserCreateDesktopEx",
        "NtUserCreateInputContext",
        "stub_UserCreateWindowStation",
        "stub_UserCtxDisplayIOCtl",
        "stub_UserDeferWindowPosAndBand",
        "NtUserDelegateCapturePointers",
        "NtUserDelegateInput",
        "NtUserDestroyDCompositionHwndTarget",
        "NtUserDestroyInputContext",
        "stub_UserDisableImmersiveOwner",
        "stub_UserDisableProcessWindowFiltering",
        "NtUserDisableThreadIme",
        "stub_UserDiscardPointerFrameMessages",
        "NtUserDisplayConfigGetDeviceInfo",
        "stub_UserDisplayConfigSetDeviceInfo",
        "stub_UserDoSoundConnect",
        "stub_UserDoSoundDisconnect",
        "stub_UserDragDetect",
        "stub_UserDragObject",
        "stub_UserDrawAnimatedRects",
        "stub_UserDrawCaption",
        "stub_UserDrawCaptionTemp",
        "stub_UserDrawMenuBarTemp",
        "stub_UserDwmGetRemoteSessionOcclusionEvent",
        "stub_UserDwmGetRemoteSessionOcclusionState",
        "stub_UserDwmKernelShutdown",
        "stub_UserDwmKernelStartup",
        "stub_UserDwmValidateWindow",
        "stub_UserEnableChildWindowDpiMessage",
        "stub_UserEnableIAMAccess",
        "NtUserEnableMouseInPointer",
        "stub_UserEnableMouseInputForCursorSuppression",
        "stub_UserEnableNonClientDpiScaling",
        "stub_UserEnableTouchPad",
        "NtUserEndMenu",
        "stub_UserEvent",
        "stub_UserFlashWindowEx",
        "stub_UserFrostCrashedWindow",
        "stub_UserGetAppImeLevel",
        "stub_UserGetAutoRotationState",
        "NtUserGetCIMSSM",
        "NtUserGetCaretPos",
        "stub_UserGetClipCursor",
        "stub_UserGetClipboardAccessToken",
        "stub_UserGetClipboardViewer",
        "stub_UserGetComboBoxInfo",
        "NtUserGetCurrentInputMessageSource",
        "NtUserGetCursorDims",
        "NtUserGetCursorInfo",
        "NtUserGetDManipHookInitFunction",
        "stub_UserGetDesktopID",
        "stub_UserGetDisplayAutoRotationPreferences",
        "stub_UserGetDisplayAutoRotationPreferencesByProcessId",
        "NtUserGetDisplayConfigBufferSizes",
        "NtUserGetDpiForMonitor",
        "NtUserGetGestureConfig",
        "stub_UserGetGestureExtArgs",
        "NtUserGetGestureInfo",
        "stub_UserGetGuiResources",
        "stub_UserGetHimetricScaleFactorFromPixelLocation",
        "NtUserGetImeHotKey",
        "NtUserGetImeInfoEx",
        "stub_UserGetInputLocaleInfo",
        "stub_UserGetInteractiveControlDeviceInfo",
        "stub_UserGetInteractiveControlInfo",
        "stub_UserGetInternalWindowPos",
        "NtUserGetKeyNameText",
        "NtUserGetKeyboardLayoutName",
        "NtUserGetLayeredWindowAttributes",
        "stub_UserGetListBoxInfo",
        "stub_UserGetMenuIndex",
        "NtUserGetMenuItemRect",
        "stub_UserGetMouseMovePointsEx",
        "NtUserGetPhysicalDeviceRect",
        "NtUserGetPointerCursorId",
        "NtUserGetPointerDevice",
        "stub_UserGetPointerDeviceCursors",
        "NtUserGetPointerDeviceProperties",
        "NtUserGetPointerDeviceRects",
        "NtUserGetPointerDevices",
        "NtUserGetPointerFrameArrivalTimes",
        "NtUserGetPointerInfoList",
        "NtUserGetPointerInputTransform",
        "NtUserGetPointerType",
        "NtUserGetPrecisionTouchPadConfiguration",
        "stub_UserGetPriorityClipboardFormat",
        "NtUserGetProcessDpiAwarenessContext",
        "NtUserGetProcessUIContextInformation",
        "NtUserGetQueueStatusReadonly",
        "stub_UserGetRawInputBuffer",
        "NtUserGetRawInputData",
        "NtUserGetRawInputDeviceInfo",
        "NtUserGetRawInputDeviceList",
        "NtUserGetRawPointerDeviceData",
        "stub_UserGetRegisteredRawInputDevices",
        "stub_UserGetTopLevelWindow",
        "NtUserGetTouchInputInfo",
        "stub_UserGetTouchValidationStatus",
        "stub_UserGetUpdatedClipboardFormats",
        "stub_UserGetWOWClass",
        "NtUserGetWindowBand",
        "NtUserGetWindowCompositionAttribute",
        "stub_UserGetWindowCompositionInfo",
        "stub_UserGetWindowDisplayAffinity",
        "NtUserGetWindowFeedbackSetting",
        "stub_UserGetWindowMinimizeRect",
        "stub_UserGetWindowRgnEx",
        "stub_UserGhostWindowFromHungWindow",
        "NtUserHandleDelegatedInput",
        "stub_UserHardErrorControl",
        "NtUserHidePointerContactVisualization",
        "stub_UserHiliteMenuItem",
        "stub_UserHungWindowFromGhostWindow",
        "stub_UserHwndQueryRedirectionInfo",
        "stub_UserHwndSetRedirectionInfo",
        "stub_UserImpersonateDdeClientWindow",
        "stub_UserInheritWindowMonitor",
        "stub_UserInitTask",
        "stub_UserInitialize",
        "stub_UserInitializeClientPfnArrays",
        "stub_UserInitializeInputDeviceInjection",
        "stub_UserInitializePointerDeviceInjection",
        "stub_UserInitializeTouchInjection",
        "stub_UserInjectDeviceInput",
        "stub_UserInjectGesture",
        "stub_UserInjectKeyboardInput",
        "stub_UserInjectMouseInput",
        "stub_UserInjectPointerInput",
        "stub_UserInjectTouchInput",
        "stub_UserInteractiveControlQueryUsage",
        "stub_UserInternalClipCursor",
        "stub_UserInternalGetWindowIcon",
        "NtUserIsChildWindowDpiMessageEnabled",
        "stub_UserIsMouseInPointerEnabled",
        "stub_UserIsMouseInputEnabled",
        "NtUserIsNonClientDpiScalingEnabled",
        "NtUserIsTopLevelWindow",
        "NtUserIsTouchWindow",
        "NtUserIsWindowBroadcastingDpiToChildren",
        "stub_UserLayoutCompleted",
        "NtUserLinkDpiCursor",
        "NtUserLoadKeyboardLayoutEx",
        "stub_UserLockWindowStation",
        "stub_UserLockWorkStation",
        "NtUserLogicalToPerMonitorDPIPhysicalPoint",
        "NtUserLogicalToPhysicalPoint",
        "stub_UserMNDragLeave",
        "stub_UserMNDragOver",
        "stub_UserMagControl",
        "stub_UserMagGetContextInformation",
        "stub_UserMagSetContextInformation",
        "NtUserMenuItemFromPoint",
        "stub_UserMinMaximize",
        "NtUserModifyWindowTouchCapability",
        "stub_UserNavigateFocus",
        "NtUserNotifyIMEStatus",
        "NtUserOpenInputDesktop",
        "stub_UserOpenThreadDesktop",
        "stub_UserPaintMonitor",
        "stub_UserPerMonitorDPIPhysicalToLogicalPoint",
        "NtUserPhysicalToLogicalPoint",
        "stub_UserPrintWindow",
        "stub_UserPromoteMouseInPointer",
        "NtUserPromotePointer",
        "stub_UserQueryBSDRWindow",
        "NtUserQueryDisplayConfig",
        "stub_UserQueryInformationThread",
        "NtUserQueryInputContext",
        "stub_UserQuerySendMessage",
        "stub_UserRealChildWindowFromPoint",
        "stub_UserRealWaitMessageEx",
        "stub_UserRegisterBSDRWindow",
        "stub_UserRegisterDManipHook",
        "stub_UserRegisterEdgy",
        "stub_UserRegisterErrorReportingDialog",
        "NtUserRegisterHotKey",
        "stub_UserRegisterManipulationThread",
        "stub_UserRegisterPointerDeviceNotifications",
        "stub_UserRegisterPointerInputTarget",
        "NtUserRegisterRawInputDevices",
        "stub_UserRegisterServicesProcess",
        "stub_UserRegisterSessionPort",
        "stub_UserRegisterShellPTPListener",
        "stub_UserRegisterTasklist",
        "NtUserRegisterTouchHitTestingWindow",
        "stub_UserRegisterTouchPadCapable",
        "stub_UserRegisterUserApiHook",
        "stub_UserReleaseDwmHitTestWaiters",
        "stub_UserRemoteConnect",
        "stub_UserRemoteRedrawRectangle",
        "stub_UserRemoteRedrawScreen",
        "stub_UserRemoteStopScreenUpdates",
        "stub_UserRemoveClipboardFormatListener",
        "stub_UserRemoveInjectionDevice",
        "NtUserReportInertia",
        "stub_UserResolveDesktopForWOW",
        "stub_UserSendEventMessage",
        "stub_UserSendInteractiveControlHapticsReport",
        "stub_UserSetActivationFilter",
        "stub_UserSetActiveProcessForMonitor",
        "stub_UserSetAppImeLevel",
        "stub_UserSetAutoRotation",
        "stub_UserSetBrokeredForeground",
        "stub_UserSetCalibrationData",
        "stub_UserSetChildWindowNoActivate",
        "stub_UserSetClassWord",
        "stub_UserSetCoreWindow",
        "stub_UserSetCoreWindowPartner",
        "stub_UserSetCursorContents",
        "stub_UserSetDisplayAutoRotationPreferences",
        "stub_UserSetDisplayConfig",
        "stub_UserSetDisplayMapping",
        "stub_UserSetFallbackForeground",
        "stub_UserSetFeatureReportResponse",
        "NtUserSetGestureConfig",
        "NtUserSetImeHotKey",
        "NtUserSetImeInfoEx",
        "NtUserSetImeOwnerWindow",
        "stub_UserSetInteractiveControlFocus",
        "stub_UserSetInteractiveCtrlRotationAngle",
        "stub_UserSetInternalWindowPos",
        "NtUserSetLayeredWindowAttributes",
        "stub_UserSetManipulationInputTarget",
        "stub_UserSetMenu",
        "stub_UserSetMenuContextHelpId",
        "stub_UserSetMenuFlagRtoL",
        "stub_UserSetMirrorRendering",
        "stub_UserSetObjectInformation",
        "stub_UserSetPrecisionTouchPadConfiguration",
        "NtUserSetProcessDpiAwarenessContext",
        "NtUserSetProcessInteractionFlags",
        "stub_UserSetProcessRestrictionExemption",
        "stub_UserSetProcessUIAccessZorder",
        "stub_UserSetShellWindowEx",
        "stub_UserSetSysColors",
        "stub_UserSetSystemCursor",
        "stub_UserSetSystemTimer",
        "NtUserSetThreadInputBlocked",
        "NtUserSetThreadLayoutHandles",
        "stub_UserSetWindowArrangement",
        "stub_UserSetWindowBand",
        "NtUserSetWindowCompositionAttribute",
        "NtUserSetWindowCompositionTransition",
        "stub_UserSetWindowDisplayAffinity",
        "NtUserSetWindowFeedbackSetting",
        "stub_UserSetWindowRgnEx",
        "stub_UserSetWindowShowState",
        "stub_UserSetWindowStationUser",
        "stub_UserShowSystemCursor",
        "stub_UserShutdownBlockReasonCreate",
        "stub_UserShutdownBlockReasonQuery",
        "stub_UserShutdownReasonDestroy",
        "stub_UserSignalRedirectionStartComplete",
        "stub_UserSlicerControl",
        "stub_UserSoundSentry",
        "stub_UserSwitchDesktop",
        "stub_UserSystemParametersInfoForDpi",
        "stub_UserTestForInteractiveUser",
        "NtUserTrackPopupMenuEx",
        "NtUserTransformPoint",
        "NtUserTransformRect",
        "NtUserUndelegateInput",
        "stub_UserUnloadKeyboardLayout",
        "stub_UserUnlockWindowStation",
        "stub_UserUnregisterHotKey",
        "stub_UserUnregisterSessionPort",
        "stub_UserUnregisterUserApiHook",
        "stub_UserUpdateDefaultDesktopThumbnail",
        "NtUserUpdateInputContext",
        "stub_UserUpdateInstance",
        "NtUserUpdateLayeredWindow",
        "stub_UserUpdatePerUserSystemParameters",
        "stub_UserUpdateWindowInputSinkHints",
        "stub_UserUpdateWindowTrackingInfo",
        "stub_UserUpdateWindowTransform",
        "stub_UserUserHandleGrantAccess",
        "stub_UserValidateHandleSecure",
        "stub_UserWaitAvailableMessageEx",
        "stub_UserWaitForInputIdle",
        "stub_UserWaitForMsgAndEvent",
        "stub_UserWaitForRedirectionStartComplete",
        "NtUserWindowFromPhysicalPoint",
        "stub_ValidateCompositionSurfaceHandle",
        "NtVisualCaptureBits",
        "NtUserSetClassLongPtr",
        "NtUserSetWindowLongPtr"
    ],
    [
        "NtAccessCheck",
        "NtWorkerFactoryWorkerReady",
        "NtAcceptConnectPort",
        "NtMapUserPhysicalPagesScatter",
        "NtWaitForSingleObject",
        "NtCallbackReturn",
        "NtReadFile",
        "NtDeviceIoControlFile",
        "NtWriteFile",
        "NtRemoveIoCompletion",
        "NtReleaseSemaphore",
        "NtReplyWaitReceivePort",
        "NtReplyPort",
        "NtSetInformationThread",
        "NtSetEvent",
        "NtClose",
        "NtQueryObject",
        "NtQueryInformationFile",
        "NtOpenKey",
        "NtEnumerateValueKey",
        "NtFindAtom",
        "NtQueryDefaultLocale",
        "NtQueryKey",
        "NtQueryValueKey",
        "NtAllocateVirtualMemory",
        "NtQueryInformationProcess",
        "NtWaitForMultipleObjects32",
        "NtWriteFileGather",
        "NtSetInformationProcess",
        "NtCreateKey",
        "NtFreeVirtualMemory",
        "NtImpersonateClientOfPort",
        "NtReleaseMutant",
        "NtQueryInformationToken",
        "NtRequestWaitReplyPort",
        "NtQueryVirtualMemory",
        "NtOpenThreadToken",
        "NtQueryInformationThread",
        "NtOpenProcess",
        "NtSetInformationFile",
        "NtMapViewOfSection",
        "NtAccessCheckAndAuditAlarm",
        "NtUnmapViewOfSection",
        "NtReplyWaitReceivePortEx",
        "NtTerminateProcess",
        "NtSetEventBoostPriority",
        "NtReadFileScatter",
        "NtOpenThreadTokenEx",
        "NtOpenProcessTokenEx",
        "NtQueryPerformanceCounter",
        "NtEnumerateKey",
        "NtOpenFile",
        "NtDelayExecution",
        "NtQueryDirectoryFile",
        "NtQuerySystemInformation",
        "NtOpenSection",
        "NtQueryTimer",
        "NtFsControlFile",
        "NtWriteVirtualMemory",
        "NtCloseObjectAuditAlarm",
        "NtDuplicateObject",
        "NtQueryAttributesFile",
        "NtClearEvent",
        "NtReadVirtualMemory",
        "NtOpenEvent",
        "NtAdjustPrivilegesToken",
        "NtDuplicateToken",
        "NtContinue",
        "NtQueryDefaultUILanguage",
        "NtQueueApcThread",
        "NtYieldExecution",
        "NtAddAtom",
        "NtCreateEvent",
        "NtQueryVolumeInformationFile",
        "NtCreateSection",
        "NtFlushBuffersFile",
        "NtApphelpCacheControl",
        "NtCreateProcessEx",
        "NtCreateThread",
        "NtIsProcessInJob",
        "NtProtectVirtualMemory",
        "NtQuerySection",
        "NtResumeThread",
        "NtTerminateThread",
        "NtReadRequestData",
        "NtCreateFile",
        "NtQueryEvent",
        "NtWriteRequestData",
        "NtOpenDirectoryObject",
        "NtAccessCheckByTypeAndAuditAlarm",
        "NtQuerySystemTime",
        "NtWaitForMultipleObjects",
        "NtSetInformationObject",
        "NtCancelIoFile",
        "NtTraceEvent",
        "NtPowerInformation",
        "NtSetValueKey",
        "NtCancelTimer",
        "NtSetTimer",
        "NtAccessCheckByType",
        "NtAccessCheckByTypeResultList",
        "NtAccessCheckByTypeResultListAndAuditAlarm",
        "NtAccessCheckByTypeResultListAndAuditAlarmByHandle",
        "NtAddAtomEx",
        "NtAddBootEntry",
        "NtAddDriverEntry",
        "NtAdjustGroupsToken",
        "NtAdjustTokenClaimsAndDeviceGroups",
        "NtAlertResumeThread",
        "NtAlertThread",
        "NtAlertThreadByThreadId",
        "NtAllocateLocallyUniqueId",
        "NtAllocateReserveObject",
        "NtAllocateUserPhysicalPages",
        "NtAllocateUuids",
        "NtAlpcAcceptConnectPort",
        "NtAlpcCancelMessage",
        "NtAlpcConnectPort",
        "NtAlpcConnectPortEx",
        "NtAlpcCreatePort",
        "NtAlpcCreatePortSection",
        "NtAlpcCreateResourceReserve",
        "NtAlpcCreateSectionView",
        "NtAlpcCreateSecurityContext",
        "NtAlpcDeletePortSection",
        "NtAlpcDeleteResourceReserve",
        "NtAlpcDeleteSectionView",
        "NtAlpcDeleteSecurityContext",
        "NtAlpcDisconnectPort",
        "NtAlpcImpersonateClientContainerOfPort",
        "NtAlpcImpersonateClientOfPort",
        "NtAlpcOpenSenderProcess",
        "NtAlpcOpenSenderThread",
        "NtAlpcQueryInformation",
        "NtAlpcQueryInformationMessage",
        "NtAlpcRevokeSecurityContext",
        "NtAlpcSendWaitReceivePort",
        "NtAlpcSetInformation",
        "NtAreMappedFilesTheSame",
        "NtAssignProcessToJobObject",
        "NtAssociateWaitCompletionPacket",
        "NtCancelIoFileEx",
        "NtCancelSynchronousIoFile",
        "NtCancelTimer2",
        "NtCancelWaitCompletionPacket",
        "NtCommitComplete",
        "NtCommitEnlistment",
        "NtCommitRegistryTransaction",
        "NtCommitTransaction",
        "NtCompactKeys",
        "NtCompareObjects",
        "NtCompareTokens",
        "ArbPreprocessEntry",
        "NtCompressKey",
        "NtConnectPort",
        "NtCreateDebugObject",
        "NtCreateDirectoryObject",
        "NtCreateDirectoryObjectEx",
        "NtCreateEnclave",
        "NtCreateEnlistment",
        "NtAdjustTokenClaimsAndDeviceGroups",
        "NtCreateIRTimer",
        "NtCreateIoCompletion",
        "NtCreateJobObject",
        "ArbAddReserved",
        "NtCreateKeyTransacted",
        "NtCreateKeyedEvent",
        "NtCreateLowBoxToken",
        "NtCreateMailslotFile",
        "NtCreateMutant",
        "NtCreateNamedPipeFile",
        "NtCreatePagingFile",
        "NtCreatePartition",
        "NtCreatePort",
        "NtCreatePrivateNamespace",
        "NtCreateProcess",
        "NtCreateProfile",
        "NtCreateProfileEx",
        "NtCreateRegistryTransaction",
        "NtCreateResourceManager",
        "NtCreateSemaphore",
        "NtCreateSymbolicLinkObject",
        "NtCreateThreadEx",
        "NtCreateTimer",
        "NtCreateTimer2",
        "NtCreateToken",
        "NtCreateTokenEx",
        "NtCreateTransaction",
        "NtCreateTransactionManager",
        "NtCreateUserProcess",
        "NtCreateWaitCompletionPacket",
        "NtCreateWaitablePort",
        "NtCreateWnfStateName",
        "NtCreateWorkerFactory",
        "NtDebugActiveProcess",
        "NtDebugContinue",
        "NtDeleteAtom",
        "NtDeleteBootEntry",
        "NtDeleteDriverEntry",
        "NtDeleteFile",
        "NtDeleteKey",
        "NtDeleteObjectAuditAlarm",
        "NtDeletePrivateNamespace",
        "NtDeleteValueKey",
        "NtDeleteWnfStateData",
        "NtDeleteWnfStateName",
        "NtDisableLastKnownGood",
        "NtDisplayString",
        "NtDrawText",
        "NtEnableLastKnownGood",
        "NtEnumerateBootEntries",
        "NtEnumerateDriverEntries",
        "NtEnumerateSystemEnvironmentValuesEx",
        "NtEnumerateTransactionObject",
        "NtExtendSection",
        "NtFilterBootOption",
        "NtFilterToken",
        "NtAdjustTokenClaimsAndDeviceGroups",
        "NtFlushBuffersFileEx",
        "NtFlushInstallUILanguage",
        "ArbPreprocessEntry",
        "NtFlushKey",
        "NtFlushProcessWriteBuffers",
        "NtFlushVirtualMemory",
        "NtFlushWriteBuffer",
        "NtFreeUserPhysicalPages",
        "NtFreezeRegistry",
        "NtFreezeTransactions",
        "NtGetCachedSigningLevel",
        "NtGetCompleteWnfStateSubscription",
        "NtGetContextThread",
        "NtGetCurrentProcessorNumber",
        "NtGetCurrentProcessorNumberEx",
        "NtGetDevicePowerState",
        "NtGetMUIRegistryInfo",
        "NtGetNextProcess",
        "NtGetNextThread",
        "NtGetNlsSectionPtr",
        "NtGetNotificationResourceManager",
        "NtGetWriteWatch",
        "NtImpersonateAnonymousToken",
        "NtImpersonateThread",
        "NtInitializeEnclave",
        "NtInitializeNlsFiles",
        "NtInitializeRegistry",
        "NtInitiatePowerAction",
        "NtIsSystemResumeAutomatic",
        "NtIsUILanguageComitted",
        "NtListenPort",
        "NtLoadDriver",
        "NtLoadEnclaveData",
        "NtLoadKey",
        "NtLoadKey2",
        "NtLoadKeyEx",
        "NtLockFile",
        "NtLockProductActivationKeys",
        "NtLockRegistryKey",
        "NtLockVirtualMemory",
        "NtMakePermanentObject",
        "NtMakeTemporaryObject",
        "NtManagePartition",
        "NtMapCMFModule",
        "NtMapUserPhysicalPages",
        "NtModifyBootEntry",
        "NtModifyDriverEntry",
        "NtNotifyChangeDirectoryFile",
        "NtNotifyChangeKey",
        "NtNotifyChangeMultipleKeys",
        "NtNotifyChangeSession",
        "NtOpenEnlistment",
        "NtAdjustTokenClaimsAndDeviceGroups",
        "NtOpenIoCompletion",
        "NtOpenJobObject",
        "NtOpenKeyEx",
        "NtOpenKeyTransacted",
        "NtOpenKeyTransactedEx",
        "NtOpenKeyedEvent",
        "NtOpenMutant",
        "NtOpenObjectAuditAlarm",
        "NtOpenPartition",
        "NtOpenPrivateNamespace",
        "NtOpenProcessToken",
        "NtOpenRegistryTransaction",
        "NtOpenResourceManager",
        "NtOpenSemaphore",
        "NtOpenSession",
        "NtOpenSymbolicLinkObject",
        "NtOpenThread",
        "NtOpenTimer",
        "NtOpenTransaction",
        "NtOpenTransactionManager",
        "NtPlugPlayControl",
        "NtPrePrepareComplete",
        "NtPrePrepareEnlistment",
        "NtPrepareComplete",
        "NtPrepareEnlistment",
        "NtPrivilegeCheck",
        "NtPrivilegeObjectAuditAlarm",
        "NtPrivilegedServiceAuditAlarm",
        "NtPropagationComplete",
        "NtPropagationFailed",
        "NtPulseEvent",
        "NtQueryBootEntryOrder",
        "NtQueryBootOptions",
        "NtQueryDebugFilterState",
        "NtQueryDirectoryObject",
        "NtQueryDriverEntryOrder",
        "NtQueryEaFile",
        "NtQueryFullAttributesFile",
        "NtQueryInformationAtom",
        "NtQueryInformationEnlistment",
        "NtQueryInformationJobObject",
        "NtQueryInformationPort",
        "NtQueryInformationResourceManager",
        "NtQueryInformationTransaction",
        "NtQueryInformationTransactionManager",
        "NtQueryInformationWorkerFactory",
        "NtQueryInstallUILanguage",
        "NtQueryIntervalProfile",
        "NtQueryIoCompletion",
        "NtQueryLicenseValue",
        "NtQueryMultipleValueKey",
        "NtQueryMutant",
        "NtQueryOpenSubKeys",
        "NtQueryOpenSubKeysEx",
        "CmpForceInvalidatePreCallback",
        "NtQueryQuotaInformationFile",
        "NtQuerySecurityAttributesToken",
        "NtQuerySecurityObject",
        "NtQuerySecurityPolicy",
        "NtQuerySemaphore",
        "NtQuerySymbolicLinkObject",
        "NtQuerySystemEnvironmentValue",
        "NtQuerySystemEnvironmentValueEx",
        "NtQuerySystemInformationEx",
        "NtQueryTimerResolution",
        "NtQueryWnfStateData",
        "NtQueryWnfStateNameInformation",
        "NtQueueApcThreadEx",
        "NtRaiseException",
        "NtRaiseHardError",
        "NtReadOnlyEnlistment",
        "NtRecoverEnlistment",
        "NtRecoverResourceManager",
        "NtRecoverTransactionManager",
        "NtRegisterProtocolAddressInformation",
        "NtRegisterThreadTerminatePort",
        "NtReleaseKeyedEvent",
        "NtReleaseWorkerFactoryWorker",
        "NtRemoveIoCompletionEx",
        "NtRemoveProcessDebug",
        "NtRenameKey",
        "NtRenameTransactionManager",
        "NtReplaceKey",
        "NtReplacePartitionUnit",
        "NtReplyWaitReplyPort",
        "NtRequestPort",
        "NtResetEvent",
        "NtResetWriteWatch",
        "NtRestoreKey",
        "NtResumeProcess",
        "NtRevertContainerImpersonation",
        "NtRollbackComplete",
        "NtRollbackEnlistment",
        "NtRollbackRegistryTransaction",
        "NtRollbackTransaction",
        "NtRollforwardTransactionManager",
        "NtSaveKey",
        "NtSaveKeyEx",
        "NtSaveMergedKeys",
        "NtSecureConnectPort",
        "NtSerializeBoot",
        "NtSetBootEntryOrder",
        "NtSetBootOptions",
        "NtSetCachedSigningLevel",
        "NtSetCachedSigningLevel2",
        "NtSetContextThread",
        "NtSetDebugFilterState",
        "NtSetDefaultHardErrorPort",
        "NtSetDefaultLocale",
        "NtSetDefaultUILanguage",
        "NtSetDriverEntryOrder",
        "NtSetEaFile",
        "NtAdjustTokenClaimsAndDeviceGroups",
        "NtAdjustTokenClaimsAndDeviceGroups",
        "NtSetIRTimer",
        "NtSetInformationDebugObject",
        "NtSetInformationEnlistment",
        "NtSetInformationJobObject",
        "NtSetInformationKey",
        "NtSetInformationResourceManager",
        "NtSetInformationSymbolicLink",
        "NtSetInformationToken",
        "NtSetInformationTransaction",
        "NtSetInformationTransactionManager",
        "NtSetInformationVirtualMemory",
        "NtSetInformationWorkerFactory",
        "NtSetIntervalProfile",
        "NtSetIoCompletion",
        "NtSetIoCompletionEx",
        "BvgaSetVirtualFrameBuffer",
        "NtAdjustTokenClaimsAndDeviceGroups",
        "NtAdjustTokenClaimsAndDeviceGroups",
        "NtSetQuotaInformationFile",
        "NtSetSecurityObject",
        "NtSetSystemEnvironmentValue",
        "NtSetSystemEnvironmentValueEx",
        "NtSetSystemInformation",
        "NtSetSystemPowerState",
        "NtSetSystemTime",
        "NtSetThreadExecutionState",
        "NtSetTimer2",
        "NtSetTimerEx",
        "NtSetTimerResolution",
        "NtSetUuidSeed",
        "NtSetVolumeInformationFile",
        "NtSetWnfProcessNotificationEvent",
        "NtShutdownSystem",
        "NtShutdownWorkerFactory",
        "NtSignalAndWaitForSingleObject",
        "NtSinglePhaseReject",
        "NtStartProfile",
        "NtStopProfile",
        "NtSubscribeWnfStateChange",
        "NtSuspendProcess",
        "NtSuspendThread",
        "NtSystemDebugControl",
        "NtTerminateJobObject",
        "NtTestAlert",
        "NtThawRegistry",
        "NtThawTransactions",
        "NtTraceControl",
        "NtTranslateFilePath",
        "NtUmsThreadYield",
        "NtUnloadDriver",
        "NtUnloadKey",
        "NtUnloadKey2",
        "NtUnloadKeyEx",
        "NtUnlockFile",
        "NtUnlockVirtualMemory",
        "NtUnmapViewOfSectionEx",
        "NtUnsubscribeWnfStateChange",
        "NtUpdateWnfStateData",
        "NtAdjustTokenClaimsAndDeviceGroups",
        "NtWaitForAlertByThreadId",
        "NtWaitForDebugEvent",
        "NtWaitForKeyedEvent",
        "NtWaitForWorkViaWorkerFactory",
        "NtAdjustTokenClaimsAndDeviceGroups",
        "NtAdjustTokenClaimsAndDeviceGroups"
    ]
]